/****************************************************************************

 Copyright (c) 1999,2000,2001 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 bbftp_utils.c v 2.0.0 2001/03/01   - Creation of the routines
               v 2.0.1 2001/04/19   - Correct indentation 


*****************************************************************************/
#include <netinet/in.h>
#include <signal.h>
#include <stdarg.h>
#include <stdio.h>

#include <bbftp.h>
#include <common.h>
#include <client.h>
#include <client_proto.h>
#include <structures.h>

extern  int     incontrolsock ;
extern  int     outcontrolsock ;
extern  int     sshchildpid  ;
extern  char    *curfilename ;
extern  char    *realfilename ;
extern  int     *myports ;
extern  int     *mysockets ;
extern  char    *readbuffer ;
extern  char    *compbuffer ;
extern  int     *mychildren ;
extern  int     debug ;
extern  int     requestedstreamnumber ;

my64_t convertlong(my64_t v) {
    struct bb {
        int    fb ;
        int sb ;
    } ;
    struct bb *bbpt ;
    int     tmp ;
    my64_t    tmp64 ;
    
    tmp64 = v ;
    bbpt = (struct bb *) &tmp64 ;
    tmp = bbpt->fb ;
    bbpt->fb = ntohl(bbpt->sb) ;
    bbpt->sb = ntohl(tmp) ;    
    return tmp64 ;
}

void printmessage(FILE *strm , int flag, int errcode, int tok, char *fmt, ...) 
{
    va_list ap;
    time_t  logtime ;
    char    clogtime[50] ;
    
    va_start(ap,fmt);
    
    /*
    ** If timestamp start to print the time
    */
    if (tok) {
        /*
        ** Get time
        */
        if ( time(&logtime) == -1 ) {
            strcpy(clogtime,"") ;
        } else {
            if ( strftime(clogtime,sizeof(clogtime),"%a %b %d %H:%M:%S (%Z)",localtime(&logtime)) == 0 )    {
                strcpy(clogtime,"") ;
            }
        }
        /*
        ** And print it
        */
        if ( strlen(clogtime) > 0 ) {
            (void) fprintf(strm,"%s : ",clogtime) ;
        }
    }
    /*
    ** Check if it is an error
    */
    if ( flag == CASE_ERROR || flag == CASE_FATAL_ERROR ) {
        /*
        ** It is an error
        */
        (void) fprintf(strm,"BBFTP-ERROR-%05d : ",errcode) ;
    } else if ( flag == CASE_WARNING ) {
        /*
        ** It is a warning
        */
        (void) fprintf(strm,"BBFTP-WARNING-%05d : ",errcode) ;
    }
    /*
    ** And print the requested string 
    */
    (void) vfprintf(strm, fmt, ap);
    if ( flag == CASE_FATAL_ERROR ) {
        /*
        ** exit in case of error
        */
        exit(errcode) ;
    }
    va_end(ap);
}

void Usage() 
{
    printf("Usage bbftp -i inputfile | -e \"command line\" \n") ;
    printf("            -u username\n") ;
    printf("           [-b (background)]\n") ;
    printf("           [-c (gzip compress)]\n") ;
    printf("           [-E server command for ssh]\n") ;
    printf("           [-f errorfile]\n") ;
    printf("           [-I ssh identity file]\n") ;
    printf("           [-L ssh command]\n") ;
    printf("           [-o outputfile]\n") ;
    printf("           [-p number of // streams]\n") ;
    printf("           [-r number of tries ]\n") ;
    printf("           [-R .bbftprc filename]\n") ;
    printf("           [-s (use ssh)]\n") ;
    printf("           [-S (use ssh in batch mode)]\n") ;
    printf("           [-t (timestamp)]\n") ;
    printf("           [-V (verbose)]\n") ;
    printf("           [-w controlport]\n") ;
    printf("           [-W (print warning to stderr) ]\n") ;
    printf("           host\n") ;
    printf("      or \n") ;
    printf("      bbftp -v\n") ;
}

/*
** Taken from Free Software Foundation
*/
void strip_trailing_slashes (char *path)
{
  int last;

  last = strlen (path) - 1;
  while (last > 0 && path[last] == '/')
    path[last--] = '\0';
}

void bbftp_close_control() 
{

    close(incontrolsock) ;
    if ( incontrolsock != outcontrolsock) close(outcontrolsock) ;
    if ( sshchildpid != 0 ) {
        kill(sshchildpid,SIGKILL) ;
    }
}
void bbftp_free_all_var() 
{

    if ( myports != NULL ) {
        free(myports) ;
        myports = NULL ;
    }
     if ( mychildren != NULL ) {
        free(mychildren) ;
        mychildren = NULL ;
    }
   if ( realfilename != NULL ) {
        free(realfilename) ;
        realfilename = NULL ;
    }
    if ( curfilename != NULL ) {
        free(curfilename) ;
        curfilename = NULL ;
    }
    if ( mysockets != NULL ) {
        free(mysockets) ;
        mysockets = NULL ;
    }
    if ( readbuffer != NULL ) {
        free(readbuffer) ;
        readbuffer = NULL ;
    }
    if ( compbuffer != NULL ) {
        free(compbuffer) ;
        compbuffer = NULL ;
    }
}
void bbftp_clean_child()
{
    int     *pidfree ;
    int     i ;
    
    if (mychildren != NULL ) {
        pidfree = mychildren ;
        for ( i=0 ; i<requestedstreamnumber ; i++) {
            if ( *pidfree != 0 ) {
                if (debug) printf("Killing child %d\n",*pidfree) ; 
                kill(*pidfree,SIGKILL) ;
            }
            *pidfree++ = 0 ;
        }
    }
}
