/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN:
        0  Keep the connection open (does not mean that the directory has been
           successfully created)
        -1 Tell the calling program to close the connection
  
 cd_v2.c v 2.0.0  2001/02/12    - Creation of the routine.
         v 2.0.1  2001/04/23    - Correct indentation
                                     
 *****************************************************************************/

#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <stdio.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

#include <bbftpd.h>
#include <common.h>
#include <structures.h>

extern int transferoption ;

int cd_v2(int sock,int msglen) {

    char    *buffer ;
    int        savederrno ;
    char    *logmessage ;
    struct  mess_dir *msg_dir ;
    char    *dirname ;

    if ( (buffer = (char *) malloc (msglen+1) ) == NULL ) {
         syslog(LOG_ERR,"Unable to malloc space for directory name (%d)",msglen) ;
         reply(MSG_BAD,"Unable to malloc space for directory name") ;
        return 0 ;
    }
    if ( (logmessage = (char *) malloc (msglen+1024) ) == NULL ) {
         syslog(LOG_ERR,"Unable to malloc space for logmessage ") ;
         reply(MSG_BAD,"Unable to malloc space for logmessage") ;
        free(buffer) ;
        return 0 ;
    }
    /*
    ** Read the characteristics of the directory
    */
    if ( readmessage(sock,buffer,msglen,CONTROLSOCKTO) < 0 ) {
        /*
        ** Error ...
        */
         syslog(LOG_ERR,"Error reading directory name") ;
        free(buffer) ;
        free(logmessage) ;
        return -1 ;
    }
     syslog(LOG_NOTICE,"in cd_v2 , msglen = %d",msglen) ;
    /*
    ** buffer contains the directory to create
    */
    buffer[msglen] = '\0' ;
    msg_dir = (struct mess_dir *) buffer ;
     transferoption  = msg_dir->transferoption ;
    dirname = msg_dir->dirname ;
    
    if ( (transferoption & TROPT_RFIO) == TROPT_RFIO ) {
        syslog(LOG_ERR,"Changing directory is not allowed under RFIO") ;
        sprintf(logmessage,"Changing directory is not allowed under RFIO") ;
        reply(MSG_BAD_NO_RETRY,logmessage) ;
        free(logmessage) ;
        free(buffer) ;
        return 0 ;
    }
    syslog(LOG_NOTICE,"Changing directory to %s",dirname) ;
    /*
    ** We change the directory
    */
    if ( chdir(dirname) < 0 ) {
        /*
        ** Depending on errno we are going to tell the client to 
        ** retry or not
        */
        savederrno = errno ;
        sprintf(logmessage,"Error changing directory %s : %s ",dirname,strerror(errno)) ;
        syslog(LOG_ERR,"Error changing directory %s : %s",dirname,strerror(errno)) ;
        /*
        ** We tell the client not to retry in the following case (even in waiting
        ** WAITRETRYTIME the problem will not be solved) :
        **        EACCES        : Search permission denied
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG: Path argument too long
        **        ENOTDIR        : A component in path is not a directory
        **        ENOENT      : A component of the path prefix does not exist or is a null pathname.
        */
        if ( savederrno == EACCES ||
                savederrno == ELOOP ||
                savederrno == ENAMETOOLONG ||
                savederrno == ENOTDIR ||
                savederrno == ENOENT ) {
            reply(MSG_BAD_NO_RETRY,logmessage) ;
            free(buffer) ;
            free(logmessage) ;
        } else {
            reply(MSG_BAD,logmessage) ;
            free(logmessage) ;
            free(buffer) ;
        }
        return 0 ;
    } else {
        if ( getcwd(logmessage,msglen+1024)  == NULL ) {
            /*
            ** Unable to get current directory
            */
            sprintf(logmessage,"Unable to get current directory") ;
            reply(MSG_BAD,logmessage) ;
            free(logmessage) ;
            free(buffer) ;
            return 0 ;
        } 
    }
    reply(MSG_OK,logmessage) ;
    free(logmessage) ;
    free(buffer) ;
    return 0 ;
}
