/****************************************************************************

 Copyright (c) 1999,2000,2001 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN:
        0  Keep the connection open (does not mean that the directory has been
           successfully created)
        -1 Tell the calling program to close the connection
  
 list_rfio_v2.c v 2.0.0  2001/02/27 - Creation of the routine.
                v 2.0.1  2001/04/23 - Correct indentation
                                    - Port to IRIX
                                      
 *****************************************************************************/
#include <dirent.h>
#include <errno.h>
#include <fnmatch.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <bbftpd.h>
#include <common.h>
#include <config.h>
#include <status.h>
#include <structures.h>

#ifdef RFIO
#include <shift.h>
#endif

extern int  transferoption ;
extern int  outcontrolsock ;

int list_rfio(char *pattern,char *logmessage) {

#ifndef RFIO
    /*
    ** Just reply that RFIO is not supported
    */
    syslog(LOG_NOTICE,"Fail to LIST : RFIO not supported") ;
    sprintf(logmessage,"Fail to LIST : RFIO not supported") ;
    reply(MSG_BAD_NO_RETRY,logmessage) ;
    return 0 ;
#else 
    int     lastslash ;
    char    *pointer ;
    char    *dirpath ;
     DIR     *curdir ;
    struct dirent *dp ;
    struct stat statbuf;
    int     lengthtosend;
    int     maxfilenamelength ;
    int     numberoffile ;
    char    *filename ;
    char    *filechar ;
    char    *filecharpoint ;
    char    send_buffer[MINMESSLEN] ;
    struct message *msg ;
    int     i ;
    char    *slash ;
    int     savederrno ;
    
    /*
    **  We are going allocate memory for the directory name
    **  We allocate stelent(pattern) + 1 char for \0 + 3 char
    **  if the directory is "." 
    */
    if ( (dirpath = (char *) malloc ( strlen(pattern) + 1 + 3 )) == NULL ) {
        syslog(LOG_ERR,"Error allocating memory for dirpath %s",strerror(errno)) ;
        sprintf(logmessage,"Error allocating memory for dirpath %s",strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        return 0 ;
    } 
    pointer = pattern ;
    slash = pattern ;
    lastslash = strlen(pattern) - 1 ;
    /*
    ** Within rfio as we have no chdir the file has to be an absolute 
    ** path 
    */
    if ( *slash == '/' ) {
        /*
        ** It is an absolute path 
        */
    } else {
        /*
        ** find the :
        */
        i = 0 ;
        while (*slash != ':' && i < strlen(pattern) ) {
            slash++;
            i++ ;
        }
        if ( i == strlen(pattern) ) {
             syslog(LOG_ERR,"Incorrect RFIO pattern : %s",pattern) ;
            sprintf(logmessage,"Incorrect RFIO pattern : %s",pattern) ;
            free(dirpath) ;
            reply(MSG_BAD_NO_RETRY,logmessage) ;
            return 0 ;
        }
    }
    while ( lastslash >= 0 && pointer[lastslash] != '/') lastslash-- ;
    if ( lastslash == -1 ) {
        /*
        ** No slash in the path, so the parre
        ** 
        */
        syslog(LOG_ERR,"Incorrect RFIO pattern : %s",pattern) ;
        sprintf(logmessage,"Incorrect RFIO pattern : %s",pattern) ;
        reply(MSG_BAD_NO_RETRY,logmessage) ;
        free(dirpath) ;
        return 0 ;
    } else if ( lastslash == 0 ) {
        /*
        ** A slash in first position so we are going to open the
        ** / directory
        */
        if ( (curdir = rfio_opendir("/")) == NULL ) {
            savederrno = errno ;
            syslog(LOG_ERR,"opendir / failed : %s ",strerror(errno)) ;
            sprintf(logmessage,"opendir / failed : %s ",strerror(errno)) ;
            if ( savederrno == EACCES ) {
                reply(MSG_BAD_NO_RETRY,logmessage) ;
            } else {
                reply(MSG_BAD,logmessage) ;
            }
            free(dirpath) ;
            return 0 ;
        }
        strcpy(dirpath,"/") ;
        pointer++ ;
    } else if ( lastslash == strlen(pointer) - 1 ) {
        /*
        ** The pattern end with a slash ..... error
        */
        syslog(LOG_ERR,"Pattern %s ends with a /",pattern) ;
        sprintf(logmessage,"Pattern %s ends with a /",pattern) ;
        reply(MSG_BAD_NO_RETRY,logmessage) ;
        free(dirpath) ;
        return 0 ;
    } else {
        pointer[lastslash] = '\0';
        /*
        ** Srip unnecessary / at the end of dirpath and reset 
        ** only one
        */
        strcpy(dirpath,pointer) ;
        strip_trailing_slashes(dirpath) ;
        dirpath[strlen(dirpath)+1] = '\0';
        dirpath[strlen(dirpath)] = '/';
        syslog(LOG_ERR,"going to open  : %s ",dirpath) ;
        if ( (curdir = rfio_opendir(dirpath)) == NULL ) {
            savederrno = errno ;
            syslog(LOG_ERR,"opendir %s failed : %s ",pointer,strerror(errno)) ;
            sprintf(logmessage,"opendir %s failed : %s ",pointer,strerror(errno)) ;
            if ( savederrno == EACCES ) {
                reply(MSG_BAD_NO_RETRY,logmessage) ;
            } else {
                reply(MSG_BAD,logmessage) ;
            }
            free(dirpath) ;
            return 0 ;
        }
        for ( i = 0 ; i <= lastslash ; i++ ) pointer++ ;
    }
    /*
    ** At this stage pointer point to the pattern and curdir
    ** is the opened directory and dirpath contain the
    ** directory name
    */
    /*
    ** As we are using the fnmatch routine we are obliged to
    ** first count the number of bytes we are sending.
    */
    lengthtosend = 0 ;
    maxfilenamelength = 0 ;
    numberoffile = 0 ;
    errno = 0 ;
    while ( (dp = rfio_readdir(curdir) ) != NULL) {
        if ( fnmatch(pointer, dp->d_name,0) == 0) {
            numberoffile++ ;
            lengthtosend = lengthtosend +  strlen(dirpath) + strlen(dp->d_name) + 1 + 3 ;
            if ( maxfilenamelength <  strlen(dp->d_name) ) maxfilenamelength = strlen(dp->d_name) ;
        }
    }
    /*
    ** Check errno in case of error during readdir
    ** for the following readir we are not going to check
    ** so that may be a cause of problem
    */
    if ( errno != 0 ) {
        syslog(LOG_ERR,"Error on readdir %s (%s) ",dirpath, strerror(errno)) ;
         sprintf(logmessage,"Error on readdir %s (%s) ",dirpath,strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        rfio_closedir(curdir) ;
        free(dirpath) ;
        return 0 ;
    }        
    /*
    ** Check if numberoffile is zero and reply now in this
    ** case 
    */
    if ( numberoffile == 0 ) {
        msg = (struct message *) send_buffer ;
        msg->code = MSG_LIST_REPL_V2 ;
        msg->msglen = 0 ;
        msg->msglen = lengthtosend ;
        if ( writemessage(outcontrolsock,send_buffer,MINMESSLEN,CONTROLSOCKTO) < 0 ) {
            syslog(LOG_ERR,"Error sending LISTREPL_V2 part 1") ;
            rfio_closedir(curdir) ;
            free(dirpath) ;
            return -1 ;
        }
        rfio_closedir(curdir) ;
        free(dirpath) ;
        return 0 ;
    }
    /*
    ** Second we are going to stat and lstat the files
    */
    if ( (filename = (char *) malloc(strlen(dirpath)+maxfilenamelength+1) ) == NULL ) {
        syslog(LOG_ERR,"Error allocating memory for filename %s",strerror(errno)) ;
        sprintf(logmessage,"Error allocating memory for filename %s",strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        rfio_closedir(curdir) ;
        free(dirpath) ;
        return 0 ;
    }
    if ( (filechar = (char *) malloc( numberoffile*3) ) == NULL ) {
        syslog(LOG_ERR,"Error allocating memory for filechar %s",strerror(errno)) ;
        sprintf(logmessage,"Error allocating memory for filechar %s",strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        rfio_closedir(curdir) ;
        free(dirpath) ;
        free(filename) ;
        return 0 ;
    }
    filecharpoint = filechar ;
    rfio_rewinddir(curdir) ;
    while ( (dp = rfio_readdir(curdir) ) != NULL) {
        if ( fnmatch(pointer, dp->d_name,0) == 0){
            sprintf(filename,"%s%s",dirpath,dp->d_name) ;
            if ( rfio_lstat(filename,&statbuf) < 0 ) {
                syslog(LOG_ERR,"Error lstating file %s : %s",filename,strerror(errno)) ;
                sprintf(logmessage,"Error lstating file %s",filename) ;
                reply(MSG_BAD,logmessage) ;
                rfio_closedir(curdir) ;
                free(dirpath) ;
                free(filename) ;
                free(filechar) ;
                return 0 ;
            }
            if ( (statbuf.st_mode & S_IFLNK) == S_IFLNK) {
                *filecharpoint++ = 'l' ;
                if ( rfio_stat(filename,&statbuf) < 0 ) {
                    /*
                    ** That means that the link refer to an unexisting file
                    */
                    *filecharpoint++ = 'u' ;
                } else {
                    if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
                        *filecharpoint++ = 'd' ;
                    } else {
                        *filecharpoint++ = 'f' ;
                    }
                }
            } else {
                *filecharpoint++ = ' ' ;
                if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
                    *filecharpoint++ = 'd' ;
                } else {
                    *filecharpoint++ = 'f' ;
                }
            }
            *filecharpoint++ = '\0' ;
        }
    }
    /*
    ** Now everything is ready so prepare the answer
    */
    msg = (struct message *) send_buffer ;
    msg->code = MSG_LIST_REPL_V2 ;
#ifndef WORDS_BIGENDIAN
    msg->msglen = ntohl(lengthtosend) ;
#else 
    msg->msglen = lengthtosend ;
#endif
    if ( writemessage(outcontrolsock,send_buffer,MINMESSLEN,CONTROLSOCKTO) < 0 ) {
        syslog(LOG_ERR,"Error sending LISTREPL_V2 part 1") ;
        rfio_closedir(curdir) ;
        free(dirpath) ;
        free(filename) ;
        free(filechar) ;
        return -1 ;
    }
    rfio_rewinddir(curdir) ;
    filecharpoint = filechar ;
    while ( (dp = rfio_readdir(curdir) ) != NULL) {
        if ( fnmatch(pointer, dp->d_name,0) == 0){
            sprintf(filename,"%s%s",dirpath,dp->d_name) ;
            /*
            ** Sending file name 
            */
            if ( writemessage(outcontrolsock,filename,strlen(filename)+1,CONTROLSOCKTO) < 0 ) {
                syslog(LOG_ERR,"Error sending filename %s",filename) ;
                rfio_closedir(curdir) ;
                free(dirpath) ;
                free(filename) ;
                free(filechar) ;
                return -1 ;
            }
            /*
            ** Sending file char 
            */
            if ( writemessage(outcontrolsock,filecharpoint,3,CONTROLSOCKTO) < 0 ) {
                syslog(LOG_ERR,"Error sending filechar for filename %s",filename) ;
                rfio_closedir(curdir) ;
                free(dirpath) ;
                free(filename) ;
                free(filechar) ;
                return -1 ;
            }
            filecharpoint++ ;
            filecharpoint++ ;
            filecharpoint++ ;
        }
    }
    rfio_closedir(curdir) ;
    free(dirpath) ;
    free(filename) ;
    free(filechar) ;
    return 0 ;
#endif
}
