/****************************************************************************

 Copyright (c) 1999,2000,2001 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN:
        0  Keep the connection open (does not mean that the directory has been
           successfully created)
        -1 Tell the calling program to close the connection
  
 list_v2.c v 2.0.0  2001/02/27  - Creation of the routine.
           v 2.0.1  2001/04/23  - Correct indentation
                                      
 *****************************************************************************/
#include <dirent.h>
#include <errno.h>
#include <fnmatch.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <bbftpd.h>
#include <common.h>
#include <status.h>
#include <structures.h>

extern int  transferoption ;
extern int  outcontrolsock ;

int list(char *pattern,char *logmessage) {

    int     lastslash ;
    char    *pointer ;
    char    *dirpath ;
     DIR     *curdir ;
#ifdef STANDART_FILE_CALL
    struct dirent *dp ;
    struct stat statbuf;
#else
#ifdef STANDART_READDIR_CALL
    struct dirent *dp ;
#else
    struct dirent64 *dp ;
#endif
    struct stat64 statbuf ;
#endif
    int     lengthtosend;
    int     maxfilenamelength ;
    int     numberoffile ;
    char    *filename ;
    char    *filechar ;
    char    *filecharpoint ;
    char    send_buffer[MINMESSLEN] ;
    struct message *msg ;
    int     i ;
    int     savederrno ;
    
    /*
    ** Check if it is a rfio list
    */
    if ( (transferoption & TROPT_RFIO) == TROPT_RFIO ) {
        return list_rfio(pattern,logmessage) ;
    }
    /*
    **  We are going allocate memory for the directory name
    **  We allocate stelent(pattern) + 1 char for \0 + 3 char
    **  if the directory is "." 
    */
    if ( (dirpath = (char *) malloc ( strlen(pattern) + 1 + 3 )) == NULL ) {
        syslog(LOG_ERR,"Error allocating memory for dirpath %s",strerror(errno)) ;
        sprintf(logmessage,"Error allocating memory for dirpath %s",strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        return 0 ;
    } 
    pointer = pattern ;
    lastslash = strlen(pointer) - 1 ;
    while ( lastslash >= 0 && pointer[lastslash] != '/') lastslash-- ;
    if ( lastslash == -1 ) {
        /*
        ** No slash in the path, so this is a pattern and we have
        ** to opendir .
        */
        if ( (curdir = opendir(".")) == NULL ) {
            savederrno = errno ;
            syslog(LOG_ERR,"opendir . failed : %s ",strerror(errno)) ;
            sprintf(logmessage,"opendir . failed : %s ",strerror(errno)) ;
            if ( savederrno == EACCES ) {
                reply(MSG_BAD_NO_RETRY,logmessage) ;
            } else {
                reply(MSG_BAD,logmessage) ;
            }
            free(dirpath) ;
            return 0 ;
        }
        strcpy(dirpath,"./") ;
    } else if ( lastslash == 0 ) {
        /*
        ** A slash in first position so we are going to open the
        ** / directory
        */
        if ( (curdir = opendir("/")) == NULL ) {
            savederrno = errno ;
            syslog(LOG_ERR,"opendir / failed : %s ",strerror(errno)) ;
            sprintf(logmessage,"opendir / failed : %s ",strerror(errno)) ;
            if ( savederrno == EACCES ) {
                reply(MSG_BAD_NO_RETRY,logmessage) ;
            } else {
                reply(MSG_BAD,logmessage) ;
            }
            free(dirpath) ;
            return 0 ;
        }
        strcpy(dirpath,"/") ;
        pointer++ ;
    } else if ( lastslash == strlen(pointer) - 1 ) {
        /*
        ** The filename end with a slash ..... error
        */
        syslog(LOG_ERR,"Pattern %s ends with a /",pattern) ;
        sprintf(logmessage,"Pattern %s ends with a /",pattern) ;
        reply(MSG_BAD_NO_RETRY,logmessage) ;
        free(dirpath) ;
        return 0 ;
    } else {
        pointer[lastslash] = '\0';
        /*
        ** Srip unnecessary / at the end of dirpath and reset 
        ** only one
        */
        strcpy(dirpath,pointer) ;
        strip_trailing_slashes(dirpath) ;
        dirpath[strlen(dirpath)+1] = '\0';
        dirpath[strlen(dirpath)] = '/';
        if ( (curdir = opendir(dirpath)) == NULL ) {
            savederrno = errno ;
            syslog(LOG_ERR,"opendir %s failed : %s ",dirpath,strerror(errno)) ;
            sprintf(logmessage,"opendir %s failed : %s ",dirpath,strerror(errno)) ;
            if ( savederrno == EACCES ) {
                reply(MSG_BAD_NO_RETRY,logmessage) ;
            } else {
                reply(MSG_BAD,logmessage) ;
            }
            free(dirpath) ;
            return 0 ;
        }
        for ( i = 0 ; i <= lastslash ; i++ ) pointer++ ;
    }
    /*
    ** At this stage pointer point to the pattern and curdir
    ** is the opened directory and dirpath contain the
    ** directory name
    */
    /*
    ** As we are using the fnmatch routine we are obliged to
    ** first count the number of bytes we are sending.
    */
    lengthtosend = 0 ;
    maxfilenamelength = 0 ;
    numberoffile = 0 ;
    errno = 0 ;
#ifdef STANDART_FILE_CALL
    while ( (dp = readdir(curdir) ) != NULL) {
#else
#ifdef STANDART_READDIR_CALL 
    while ( (dp = readdir(curdir) ) != NULL) {
#else
    while ( (dp = readdir64(curdir) ) != NULL) {
#endif
#endif
        if ( fnmatch(pointer, dp->d_name,0) == 0) {
            numberoffile++ ;
            lengthtosend = lengthtosend +  strlen(dirpath) + strlen(dp->d_name) + 1 + 3 ;
            if ( maxfilenamelength <  strlen(dp->d_name) ) maxfilenamelength = strlen(dp->d_name) ;
        }
    }
    /*
    ** Check errno in case of error during readdir
    ** for the following readir we are not going to check
    ** so that may be a cause of problem
    */
    if ( errno != 0 ) {
        syslog(LOG_ERR,"Error on readdir %s (%s) ",dirpath, strerror(errno)) ;
         sprintf(logmessage,"Error on readdir %s (%s) ",dirpath,strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        closedir(curdir) ;
        free(dirpath) ;
        return 0 ;
    }
    /*
    ** Check if numberoffile is zero and reply now in this
    ** case 
    */
    if ( numberoffile == 0 ) {
        msg = (struct message *) send_buffer ;
        msg->code = MSG_LIST_REPL_V2 ;
        msg->msglen = 0 ;
        msg->msglen = lengthtosend ;
        if ( writemessage(outcontrolsock,send_buffer,MINMESSLEN,CONTROLSOCKTO) < 0 ) {
            syslog(LOG_ERR,"Error sending LISTREPL_V2 part 1") ;
            closedir(curdir) ;
            free(dirpath) ;
            return -1 ;
        }
        closedir(curdir) ;
        free(dirpath) ;
        return 0 ;
    }
    /*
    ** Second we are going to stat and lstat the files
    */
    if ( (filename = (char *) malloc(strlen(dirpath)+maxfilenamelength+1) ) == NULL ) {
        syslog(LOG_ERR,"Error allocating memory for filename %s",strerror(errno)) ;
        sprintf(logmessage,"Error allocating memory for filename %s",strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        closedir(curdir) ;
        free(dirpath) ;
        return 0 ;
    }
    if ( (filechar = (char *) malloc( numberoffile*3) ) == NULL ) {
        syslog(LOG_ERR,"Error allocating memory for filechar %s",strerror(errno)) ;
        sprintf(logmessage,"Error allocating memory for filechar %s",strerror(errno)) ;
        reply(MSG_BAD,logmessage) ;
        closedir(curdir) ;
        free(dirpath) ;
        free(filename) ;
        return 0 ;
    }
    filecharpoint = filechar ;
    rewinddir(curdir) ;
#ifdef STANDART_FILE_CALL
    while ( (dp = readdir(curdir) ) != NULL) {
#else
#ifdef STANDART_READDIR_CALL 
    while ( (dp = readdir(curdir) ) != NULL) {
#else
    while ( (dp = readdir64(curdir) ) != NULL) {
#endif
#endif
        if ( fnmatch(pointer, dp->d_name,0) == 0){
            sprintf(filename,"%s%s",dirpath,dp->d_name) ;
#ifdef STANDART_FILE_CALL
            if ( lstat(filename,&statbuf) < 0 ) {
#else
            if ( lstat64(filename,&statbuf) < 0 ) {
#endif
                syslog(LOG_ERR,"Error lstating file %s : %s",filename,strerror(errno)) ;
                sprintf(logmessage,"Error lstating file %s",filename) ;
                reply(MSG_BAD,logmessage) ;
                closedir(curdir) ;
                free(dirpath) ;
                free(filename) ;
                free(filechar) ;
                return 0 ;
            }
            if ( (statbuf.st_mode & S_IFLNK) == S_IFLNK) {
                *filecharpoint++ = 'l' ;
#ifdef STANDART_FILE_CALL
                if ( stat(filename,&statbuf) < 0 ) {
#else
                if ( stat64(filename,&statbuf) < 0 ) {
#endif
                    /*
                    ** That means that the link refer to an unexisting file
                    */
                    *filecharpoint++ = 'u' ;
                } else {
                    if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
                        *filecharpoint++ = 'd' ;
                    } else {
                        *filecharpoint++ = 'f' ;
                    }
                }
            } else {
                *filecharpoint++ = ' ' ;
                if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
                    *filecharpoint++ = 'd' ;
                } else {
                    *filecharpoint++ = 'f' ;
                }
            }
            *filecharpoint++ = '\0' ;
        }
    }
    /*
    ** Now everything is ready so prepare the answer
    */
    msg = (struct message *) send_buffer ;
    msg->code = MSG_LIST_REPL_V2 ;
#ifndef WORDS_BIGENDIAN
    msg->msglen = ntohl(lengthtosend) ;
#else 
    msg->msglen = lengthtosend ;
#endif
    if ( writemessage(outcontrolsock,send_buffer,MINMESSLEN,CONTROLSOCKTO) < 0 ) {
        syslog(LOG_ERR,"Error sending LISTREPL_V2 part 1") ;
        closedir(curdir) ;
        free(dirpath) ;
        free(filename) ;
        free(filechar) ;
        return -1 ;
    }
    rewinddir(curdir) ;
    filecharpoint = filechar ;
#ifdef STANDART_FILE_CALL
    while ( (dp = readdir(curdir) ) != NULL) {
#else
#ifdef STANDART_READDIR_CALL 
    while ( (dp = readdir(curdir) ) != NULL) {
#else
    while ( (dp = readdir64(curdir) ) != NULL) {
#endif
#endif
        if ( fnmatch(pointer, dp->d_name,0) == 0){
            sprintf(filename,"%s%s",dirpath,dp->d_name) ;
            /*
            ** Sending file name 
            */
            if ( writemessage(outcontrolsock,filename,strlen(filename)+1,CONTROLSOCKTO) < 0 ) {
                syslog(LOG_ERR,"Error sending filename %s",filename) ;
                closedir(curdir) ;
                free(dirpath) ;
                free(filename) ;
                free(filechar) ;
                return -1 ;
            }
            /*
            ** Sending file char 
            */
            if ( writemessage(outcontrolsock,filecharpoint,3,CONTROLSOCKTO) < 0 ) {
                syslog(LOG_ERR,"Error sending filechar for filename %s",filename) ;
                closedir(curdir) ;
                free(dirpath) ;
                free(filename) ;
                free(filechar) ;
                return -1 ;
            }
            filecharpoint++ ;
            filecharpoint++ ;
            filecharpoint++ ;
        }
    }
    closedir(curdir) ;
    free(dirpath) ;
    free(filename) ;
    free(filechar) ;
    return 0 ;
}
