/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
 RETURN:
         0  calling program return to main loop
        -1 calling program exit
 
 readcontrol_v2.c v 2.0.0 2000/12/19    - Routine creation
                  v 2.0.1 2001/04/23    - Correct indentation
               
*****************************************************************************/
#include <errno.h>
#include <netinet/in.h>
#include <stdio.h>
#include <syslog.h>

#include <bbftpd.h>
#include <common.h>
#include <status.h>
#include <structures.h>
#include <version.h>

extern int  state ;
extern int  incontrolsock ;
extern int  outcontrolsock ;
extern char *curfilename ;
extern char *realfilename ;
extern int  curfilenamelen ;
extern int  transferoption ;
extern int  filemode ;
extern char lastaccess[9] ;
extern char lastmodif[9] ;
extern int  sendwinsize ;        
extern int  recvwinsize ;        
extern int  buffersizeperstream ;
extern int  requestedstreamnumber ;
extern my64_t  filesize ;
extern int  *myports ;
extern int  *mychildren ;
extern int  myumask ;
extern int  mycos ;
extern char *readbuffer ;
extern char *compbuffer ;

#ifndef WORDS_BIGENDIAN
my64_t    convertlong(my64_t v) ;
#endif

int readcontrol_v2(int msgcode,int msglen) {

    int        retcode ;
    int        i ;
    char    *receive_buffer ;
    int     oldumask, newumask ;
    struct  message *msg ;
    struct  mess_store_v2 *msg_store_v2 ;
    struct  mess_integer *msg_integer ;
    struct  mess_dir *msg_dir ;
    char    *logmessage ;
    int     *portfree ;
    
    /*
    ** Depending on the state the accepted message may be different
    */
#ifndef WORDS_BIGENDIAN
    msglen = ntohl(msglen) ;
#endif
    switch (state) {
        /*
        ** Case S_LOGGED :
        **      Nothing is going on, so message accepted are :
        **
        **         MSG_CHDIR_V2
        **         MSG_CHUMASK
        **         MSG_LIST_V2
        **         MSG_MKDIR_V2
        **         MSG_RETR_V2
        **         MSG_STORE_V2
        **
        */
           case S_LOGGED : {
            switch (msgcode) {
            
                case MSG_CHCOS : {
                    syslog(LOG_NOTICE,"Receiving MSG_CHCOS, msglen = %d",msglen) ;
                    if ( (receive_buffer = (char *) malloc (msglen)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for MSG_CHCOS : %s",strerror(errno)) ;
                        reply(MSG_BAD,"Error allocating memory for MSG_CHCOS") ;
                        return 0 ;
                    }
                    if ( readmessage(incontrolsock,receive_buffer,msglen,CONTROLSOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading MSG_CHCOS") ;
                        free(receive_buffer) ;
                        return -1 ;
                    }
                    msg_integer = (struct  mess_integer *) receive_buffer ;
#ifndef WORDS_BIGENDIAN
                    mycos = ntohl(msg_integer->myint) ;
#else
                    mycos = msg_integer->myint ;
#endif
                    syslog(LOG_NOTICE,"Cos set to %d",mycos) ;
                    reply(MSG_OK,"COS set") ;
                    free(receive_buffer) ;
                    return 0 ;
               }

                case MSG_CHDIR_V2 : {
                    syslog(LOG_NOTICE,"Receiving MSG_CHDIR_V2, msglen = %d",msglen) ;
                    retcode = cd_v2(incontrolsock,msglen) ;
                    return retcode ;
                }
                
                case MSG_CHUMASK : {
                    syslog(LOG_NOTICE,"Receiving MSG_UMASK ") ;
                    if ( (receive_buffer = (char *) malloc (msglen)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for MSG_UMASK : %s",strerror(errno)) ;
                        reply(MSG_BAD,"Error allocating memory for MSG_UMASK") ;
                        return 0 ;
                    }
                    if ( readmessage(incontrolsock,receive_buffer,msglen,CONTROLSOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading MSG_UMASK") ;
                        free(receive_buffer) ;
                        return -1 ;
                    }
                    msg_integer = (struct  mess_integer *) receive_buffer ;
#ifndef WORDS_BIGENDIAN
                    myumask = ntohl(msg_integer->myint) ;
#else
                    myumask = msg_integer->myint ;
#endif
                    /*
                    ** We reset the umask first then call twice the umask
                    ** in order to see the real umask set
                    */
                    oldumask = umask(0) ;
                    umask(myumask) ;
                    newumask = umask(myumask) ;
                    if ( (logmessage = (char *) malloc ( 1024 )) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for logmessage") ;
                        reply(MSG_OK,"umask changed") ;
                        free(receive_buffer) ;
                        return 0 ;
                    }
                    sprintf(logmessage,"umask changed from %04o to %04o",oldumask,newumask) ;
                    reply(MSG_OK,logmessage) ;
                    free(receive_buffer) ;
                    free(logmessage) ;
                    return 0 ;
                }
                
                case MSG_LIST_V2 :{
                   syslog(LOG_NOTICE,"Receiving MSG_LIST_V2 ") ;
                   if ( (receive_buffer = (char *) malloc (msglen+1) ) == NULL ) {
                        syslog(LOG_ERR,"Unable to malloc space for directory name (%d)",msglen) ;
                        reply(MSG_BAD,"Unable to malloc space for directory name") ;
                       return 0 ;
                   }
                   if ( (logmessage = (char *) malloc (msglen+1024) ) == NULL ) {
                        syslog(LOG_ERR,"Unable to malloc space for logmessage ") ;
                        reply(MSG_BAD,"Unable to malloc space for logmessage") ;
                       free(receive_buffer) ;
                       return 0 ;
                   }
                   /*
                   ** Read the characteristics of the directory
                   */
                   if ( readmessage(incontrolsock,receive_buffer,msglen,CONTROLSOCKTO) < 0 ) {
                       /*
                       ** Error ...
                       */
                        syslog(LOG_ERR,"Error reading directory name") ;
                       free(receive_buffer) ;
                       free(logmessage) ;
                       return -1 ;
                   }
                   /*
                   ** buffer contains the directory to create
                   */
                   receive_buffer[msglen] = '\0' ;
                   msg_dir = (struct mess_dir *) receive_buffer ;
                    transferoption  = msg_dir->transferoption ;
                   syslog(LOG_ERR,"Pattern = %s",msg_dir->dirname) ;
                   retcode = list(msg_dir->dirname,logmessage) ;
                   free(receive_buffer) ;
                   free(logmessage) ;
                   return retcode ;
                }
                
                case MSG_MKDIR_V2 : {
                    syslog(LOG_NOTICE,"Receiving MSG_MKDIR_V2 ") ;
                    retcode = mkdir_v2(incontrolsock,msglen) ;
                    return retcode ;
                }
                
                    
                case MSG_RETR_V2 : {
                    syslog(LOG_NOTICE,"Receiving MSG_RETR_V2") ;
                    if ( (receive_buffer = (char *) malloc (msglen)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for MSG_RETR_V2 : %s",strerror(errno)) ;
                        return -1 ;
                    }
                     if ( readmessage(incontrolsock,receive_buffer,msglen,CONTROLSOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading MSG_RETR_V2") ;
                        free(receive_buffer) ;
                        return -1 ;
                    }
                    msg_store_v2 = (struct  mess_store_v2 *) receive_buffer ;
                    /*
                    ** Set up the transfer parameter, we are going only to store 
                    ** them 
                    */
#ifndef WORDS_BIGENDIAN
                    transferoption          = msg_store_v2->transferoption ;
                    sendwinsize             = ntohl(msg_store_v2->sendwinsize) ;
                    recvwinsize             = ntohl(msg_store_v2->recvwinsize) ;
                    buffersizeperstream     = ntohl(msg_store_v2->buffersizeperstream) ;
                    requestedstreamnumber   = ntohl(msg_store_v2->nbstream) ;
#else
                    transferoption          = msg_store_v2->transferoption ;
                    sendwinsize             = msg_store_v2->sendwinsize ;
                    recvwinsize             = msg_store_v2->recvwinsize ;
                    buffersizeperstream     = msg_store_v2->buffersizeperstream ;
                    requestedstreamnumber   = msg_store_v2->nbstream ;
#endif
                    syslog(LOG_ERR,"transferoption         = %d ",transferoption) ;
                    syslog(LOG_ERR,"sendwinsize            = %d ",sendwinsize) ;
                    syslog(LOG_ERR,"recvwinsize            = %d ",recvwinsize) ;
                    syslog(LOG_ERR,"buffersizeperstream    = %d ",buffersizeperstream) ;
                    syslog(LOG_ERR,"requestedstreamnumber  = %d ",requestedstreamnumber) ;
                    state = S_WAITING_FILENAME_RETR ;
                    free(receive_buffer) ;
                    return 0 ;
                }
                
                case MSG_STORE_V2 :{
                    syslog(LOG_NOTICE,"Receiving MSG_STORE_V2") ;
                    if ( (receive_buffer = (char *) malloc (msglen)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for MSG_STORE_V2 : %s",strerror(errno)) ;
                        return -1 ;
                    }
                     if ( readmessage(incontrolsock,receive_buffer,msglen,CONTROLSOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading MSG_STORE_V2") ;
                        free(receive_buffer) ;
                        return -1 ;
                    }
                    msg_store_v2 = (struct  mess_store_v2 *) receive_buffer ;
                    /*
                    ** Set up the transfer parameter, we are going only to store 
                    ** them 
                    */
#ifndef WORDS_BIGENDIAN
                    transferoption          = msg_store_v2->transferoption ;
                    filemode                = ntohl(msg_store_v2->filemode) ;
                    strncpy(lastaccess,msg_store_v2->lastaccess,8) ;
                    lastaccess[8] = '\0' ;
                    strncpy(lastmodif,msg_store_v2->lastmodif,8) ;
                    lastmodif[8] = '\0' ;
                    sendwinsize             = ntohl(msg_store_v2->sendwinsize) ;
                    recvwinsize             = ntohl(msg_store_v2->recvwinsize) ;
                    buffersizeperstream     = ntohl(msg_store_v2->buffersizeperstream) ;
                    requestedstreamnumber   = ntohl(msg_store_v2->nbstream) ;
                    filesize                = convertlong(msg_store_v2->filesize) ;
#else
                    transferoption          = msg_store_v2->transferoption ;
                    filemode                = msg_store_v2->filemode ;
                    strncpy(lastaccess,msg_store_v2->lastaccess,8) ;
                    lastaccess[8] = '\0' ;
                    strncpy(lastmodif,msg_store_v2->lastmodif,8) ;
                    lastmodif[8] = '\0' ;
                    sendwinsize             = msg_store_v2->sendwinsize ;
                    recvwinsize             = msg_store_v2->recvwinsize ;
                    buffersizeperstream     = msg_store_v2->buffersizeperstream ;
                    requestedstreamnumber   = msg_store_v2->nbstream ;
                    filesize                = msg_store_v2->filesize ;
#endif
                    syslog(LOG_ERR,"transferoption         = %d ",transferoption) ;
                    syslog(LOG_ERR,"filemode               = %d ",filemode) ;
                    syslog(LOG_ERR,"lastaccess             = %s ",lastaccess) ;
                    syslog(LOG_ERR,"lastmodif              = %s ",lastmodif) ;
                    syslog(LOG_ERR,"sendwinsize            = %d ",sendwinsize) ;
                    syslog(LOG_ERR,"recvwinsize            = %d ",recvwinsize) ;
                    syslog(LOG_ERR,"buffersizeperstream    = %d ",buffersizeperstream) ;
                    syslog(LOG_ERR,"requestedstreamnumber  = %d ",requestedstreamnumber) ;
                    syslog(LOG_ERR,"filesize               = %" LONG_LONG_FORMAT " ",filesize) ;
                    state = S_WAITING_FILENAME_STORE ;
                    free(receive_buffer) ;
                    return 0 ;
                }
                
                default :{
                    syslog(LOG_ERR,"Unkown message in logged state %d",msgcode) ;
                    retcode = discardmessage(incontrolsock,msglen,CONTROLSOCKTO) ;
                    reply(MSG_BAD_NO_RETRY,"Unkown message in S_LOGGED state") ;
                    return retcode ;
                }
            }
        }
        /*
        ** Case S_WAITING_FILENAME_STORE
        **
        **         MSG_FILENAME
        */
           case S_WAITING_FILENAME_STORE : {
            switch (msgcode) {
                case MSG_FILENAME :{
                    syslog(LOG_NOTICE,"Receiving MSG_FILENAME in S_WAITING_FILENAME_STORE state") ;
                    if ( msglen == 0 ) {
                         syslog(LOG_ERR,"Filename length null") ;
                         return -1 ;
                    }
                    if ( (curfilename = (char *) malloc (msglen+1)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for filename : %s",strerror(errno)) ;
                        return -1 ;
                    }
                    if ( (realfilename = (char *) malloc (msglen+11)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for realfilename : %s",strerror(errno)) ;
                        free_all_var() ;
                        return -1 ;
                    }
                    if ( readmessage(incontrolsock,curfilename,msglen,CONTROLSOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading filename") ;
                        free_all_var() ;
                        return -1 ;
                    }
                    curfilename[msglen] = '\0' ;
                    curfilenamelen = msglen ;
                    syslog(LOG_NOTICE,"Request to store file %s",curfilename) ;
                    /*
                    ** Alloc space for logmessage 
                    */
                    if ( (logmessage = (char *) malloc ( 1024 )) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for logmessage") ;
                        free_all_var() ;
                        return -1 ;
                    }
                    if ( (retcode = storecheckoptions(logmessage)) < 0 ) {
                        reply(MSG_BAD_NO_RETRY,logmessage) ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    }
                    /*
                    ** Options are corrects
                    */
                    if ( (transferoption & TROPT_TMP ) == TROPT_TMP ) {
                        /*
                        ** Caculate tempfile name if TROPT_TMP is set
                        */
                        sprintf(realfilename,"%s.bbftp.tmp",curfilename) ;
                    } else {
                        sprintf(realfilename,"%s",curfilename) ;
                    }
                    /*
                    ** Create the file
                    */
                    if ( retcode = storecreatefile(realfilename,logmessage) < 0 ) {
                        reply(MSG_BAD_NO_RETRY,logmessage) ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    } else if ( retcode > 0 ) {
                        reply(MSG_BAD,logmessage) ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    } else {
                        if ( filesize == 0 ) {
                            if ( (transferoption & TROPT_TMP ) == TROPT_TMP ) {
                                if ( storerename(realfilename,curfilename,logmessage) < 0 ) {
                                    storeunlink(realfilename) ;
                                    reply(MSG_BAD,logmessage) ;
                                    free_all_var() ;
                                    free(logmessage) ;
                                    state = S_LOGGED ;
                                    return 0 ;
                                } else {
                                    reply(MSG_OK,"") ;
                                    free_all_var() ;
                                    free(logmessage) ;
                                    return 0 ;
                               }
                            } else {
                                reply(MSG_OK,"") ;
                                free_all_var() ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                        } else {
                            /*
                            ** We are reserving the memory for the ports
                            */
                            if ( (myports = (int *) malloc (requestedstreamnumber*sizeof(int))) == NULL ) {
                                storeunlink(realfilename) ;
                                free_all_var() ;
                                syslog(LOG_ERR,"Unable to allocate memory for myports") ;
                                sprintf(logmessage,"Unable to allocate memory for myports") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            /*
                            ** We are reserving the memory for the children pid
                            */
                            if ( (mychildren = (int *) malloc (requestedstreamnumber*sizeof(int))) == NULL) {
                                storeunlink(realfilename) ;
                                free_all_var() ;
                                syslog(LOG_ERR,"Unable to allocate memory for mychildren") ;
                                sprintf(logmessage,"Unable to allocate memory for mychildren") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            /*
                            ** We are reserving the memory for the readbuffer
                            */
                            if ( (readbuffer = (char *) malloc (buffersizeperstream*1024)) == NULL) {
                                storeunlink(realfilename) ;
                                free_all_var() ;
                                syslog(LOG_ERR,"Unable to allocate memory for readbuffer") ;
                                sprintf(logmessage,"Unable to allocate memory for readbuffer") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            /*
                            ** We are reserving the memory for the compression buffer 
                            ** if needed
                            */
                            if ( (transferoption & TROPT_GZIP ) == TROPT_GZIP ) {
                                if ( (compbuffer = (char *) malloc (buffersizeperstream*1024)) == NULL) {
                                    storeunlink(realfilename) ;
                                    free_all_var() ;
                                    syslog(LOG_ERR,"Unable to allocate memory for compbuffer") ;
                                    sprintf(logmessage,"Unable to allocate memory for compbuffer") ;
                                    reply(MSG_BAD,logmessage) ;
                                    free(logmessage) ;
                                    state = S_LOGGED ;
                                    return 0 ;
                                }
                            }
                            if ( (receive_buffer = (char *) malloc (STORMESSLEN_V2)) == NULL ) {
                                storeunlink(realfilename) ;
                                syslog(LOG_ERR,"Error allocating memory for MSG_TRANS_OK_V2 : %s",strerror(errno)) ;
                                free_all_var() ;
                                sprintf(logmessage,"Unable to allocate memory for message MSG_TRANS_OK_V2") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            msg = (struct message *) receive_buffer ;
                            msg->code = MSG_TRANS_OK_V2 ;
#ifndef WORDS_BIGENDIAN
                            msg->msglen = ntohl(STORMESSLEN_V2) ;
#else
                            msg->msglen = STORMESSLEN_V2 ;
#endif
                            if ( writemessage(outcontrolsock,receive_buffer,MINMESSLEN,SENDCONTROLTO) < 0 ) {
                                syslog(LOG_ERR,"Error wtiting MSG_TRANS_OK_V2 first part") ;
                                storeunlink(realfilename) ;
                                free_all_var() ;
                                free(logmessage) ;
                                free(receive_buffer) ;
                                return -1 ;
                            }
                            msg_store_v2 = (struct  mess_store_v2 *) receive_buffer ;
                            /*
                            ** Set up the transfer parameter
                            */
#ifndef WORDS_BIGENDIAN
                            msg_store_v2->transferoption       = transferoption ;
                            msg_store_v2->filemode             = ntohl(filemode) ;
                            strncpy(msg_store_v2->lastaccess,lastaccess,8) ;
                            msg_store_v2->lastaccess[8] = '0' ;
                            strncpy(msg_store_v2->lastmodif,lastmodif,8) ;
                            msg_store_v2->lastmodif[8] = '0' ;
                            msg_store_v2->sendwinsize          = ntohl(sendwinsize) ;
                            msg_store_v2->recvwinsize          = ntohl(recvwinsize) ;
                            msg_store_v2->buffersizeperstream  = ntohl(buffersizeperstream) ;
                            msg_store_v2->nbstream             = ntohl(requestedstreamnumber) ;
                            msg_store_v2->filesize             = convertlong(filesize) ;
#else
                            msg_store_v2->transferoption       = transferoption  ;       
                            msg_store_v2->filemode             =  filemode;
                            strncpy(msg_store_v2->lastaccess,lastaccess,8) ;
                            msg_store_v2->lastaccess[8] = '0' ;
                            strncpy(msg_store_v2->lastmodif,lastmodif,8) ;
                            msg_store_v2->lastmodif[8] = '0' ;
                            msg_store_v2->sendwinsize          = sendwinsize ;
                            msg_store_v2->recvwinsize          = recvwinsize ;
                            msg_store_v2->buffersizeperstream  = buffersizeperstream ;
                            msg_store_v2->nbstream             = requestedstreamnumber;
                            msg_store_v2->filesize             = filesize ;
#endif
                            if ( writemessage(outcontrolsock,receive_buffer,STORMESSLEN_V2,SENDCONTROLTO) < 0 ) {
                                syslog(LOG_ERR,"Error wtiting MSG_TRANS_OK_V2 second part") ;
                                storeunlink(realfilename) ;
                                free_all_var() ;
                                free(logmessage) ;
                                free(receive_buffer) ;
                                return -1 ;
                            }
                            free(logmessage) ;
                            free(receive_buffer) ;
                            state = S_WAITING_STORE_START ;
                            return 0 ;
                        }
                    }
                }
                default :{
                    syslog(LOG_ERR,"Unkown message in S_WAITING_FILENAME_STORE state : %d",msgcode) ;
                    return -1 ;
                }
            }
        }
        /*
        ** Case S_WAITING_FILENAME_RETR
        **
        **         MSG_FILENAME
        */
           case S_WAITING_FILENAME_RETR : {
            switch (msgcode) {
                case MSG_FILENAME :{
                    syslog(LOG_NOTICE,"Receiving MSG_FILENAME in S_WAITING_FILENAME_RETR state") ;
                    if ( msglen == 0 ) {
                         syslog(LOG_ERR,"Filename length null") ;
                         return -1 ;
                    }
                    if ( (curfilename = (char *) malloc (msglen+1)) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for filename : %s",strerror(errno)) ;
                        return -1 ;
                    }
                    if ( readmessage(incontrolsock,curfilename,msglen,CONTROLSOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading filename") ;
                        free_all_var() ;
                        return -1 ;
                    }
                    curfilename[msglen] = '\0' ;
                    curfilenamelen = msglen ;
                    syslog(LOG_NOTICE,"Request to retreive file %s",curfilename) ;
                    /*
                    ** Alloc space for logmessage 
                    */
                    if ( (logmessage = (char *) malloc ( 1024 )) == NULL ) {
                        syslog(LOG_ERR,"Error allocating memory for logmessage") ;
                        free_all_var() ;
                        return -1 ;
                    }
                    /*
                    ** Create the file
                    */
                    if ( retcode = retrcheckfile(curfilename,logmessage) < 0 ) {
                        reply(MSG_BAD_NO_RETRY,logmessage) ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    } else if ( retcode > 0 ) {
                        reply(MSG_BAD,logmessage) ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    } else {
                        if ( filesize != 0 ) {
                            /*
                            ** We are reserving the memory for the ports
                            */
                            if ( (myports = (int *) malloc (requestedstreamnumber*sizeof(int))) == NULL ) {
                                free_all_var() ;
                                syslog(LOG_ERR,"Unable to allocate memory for myports") ;
                                sprintf(logmessage,"Unable to allocate memory for myports") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            /*
                            ** We are reserving the memory for the children pid
                            */
                            if ( (mychildren = (int *) malloc (requestedstreamnumber*sizeof(int))) == NULL) {
                                free_all_var() ;
                                syslog(LOG_ERR,"Unable to allocate memory for mychildren") ;
                                sprintf(logmessage,"Unable to allocate memory for mychildren") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            /*
                            ** We are reserving the memory for the readbuffer
                            */
                            if ( (readbuffer = (char *) malloc (buffersizeperstream*1024)) == NULL) {
                                free_all_var() ;
                                syslog(LOG_ERR,"Unable to allocate memory for readbuffer") ;
                                sprintf(logmessage,"Unable to allocate memory for readbuffer") ;
                                reply(MSG_BAD,logmessage) ;
                                free(logmessage) ;
                                state = S_LOGGED ;
                                return 0 ;
                            }
                            /*
                            ** We are reserving the memory for the compression buffer 
                            ** if needed
                            */
                            if ( (transferoption & TROPT_GZIP ) == TROPT_GZIP ) {
                                if ( (compbuffer = (char *) malloc (buffersizeperstream*1024)) == NULL) {
                                    free_all_var() ;
                                    syslog(LOG_ERR,"Unable to allocate memory for compbuffer") ;
                                    sprintf(logmessage,"Unable to allocate memory for compbuffer") ;
                                    reply(MSG_BAD,logmessage) ;
                                    free(logmessage) ;
                                    state = S_LOGGED ;
                                    return 0 ;
                                }
                            }
                        }
                        if ( (receive_buffer = (char *) malloc (STORMESSLEN_V2)) == NULL ) {
                            syslog(LOG_ERR,"Error allocating memory for MSG_TRANS_OK_V2 : %s",strerror(errno)) ;
                            free_all_var() ;
                            sprintf(logmessage,"Unable to allocate memory for message MSG_TRANS_OK_V2") ;
                            reply(MSG_BAD,logmessage) ;
                            free(logmessage) ;
                            state = S_LOGGED ;
                            return 0 ;
                        }
                        msg = (struct message *) receive_buffer ;
                        msg->code = MSG_TRANS_OK_V2 ;
#ifndef WORDS_BIGENDIAN
                        msg->msglen = ntohl(STORMESSLEN_V2) ;
#else
                        msg->msglen = STORMESSLEN_V2 ;
#endif
                        if ( writemessage(outcontrolsock,receive_buffer,MINMESSLEN,SENDCONTROLTO) < 0 ) {
                            syslog(LOG_ERR,"Error wtiting MSG_TRANS_OK_V2 first part") ;
                            free_all_var() ;
                            free(logmessage) ;
                            free(receive_buffer) ;
                            return -1 ;
                        }
                        msg_store_v2 = (struct  mess_store_v2 *) receive_buffer ;
                        /*
                        ** Set up the transfer parameter
                        */
#ifndef WORDS_BIGENDIAN
                        msg_store_v2->transferoption       = transferoption ;
                        msg_store_v2->filemode             = ntohl(filemode) ;
                        strncpy(msg_store_v2->lastaccess,lastaccess,8) ;
                        msg_store_v2->lastaccess[8] = '0' ;
                        strncpy(msg_store_v2->lastmodif,lastmodif,8) ;
                        msg_store_v2->lastmodif[8] = '0' ;
                        msg_store_v2->sendwinsize          = ntohl(sendwinsize) ;
                        msg_store_v2->recvwinsize          = ntohl(recvwinsize) ;
                        msg_store_v2->buffersizeperstream  = ntohl(buffersizeperstream) ;
                        msg_store_v2->nbstream             = ntohl(requestedstreamnumber) ;
                        msg_store_v2->filesize             = convertlong(filesize) ;
#else
                        msg_store_v2->transferoption       = transferoption  ;       
                        msg_store_v2->filemode             =  filemode;
                        strncpy(msg_store_v2->lastaccess,lastaccess,8) ;
                        msg_store_v2->lastaccess[8] = '0' ;
                        strncpy(msg_store_v2->lastmodif,lastmodif,8) ;
                        msg_store_v2->lastmodif[8] = '0' ;
                        msg_store_v2->sendwinsize          = sendwinsize ;
                        msg_store_v2->recvwinsize          = recvwinsize ;
                        msg_store_v2->buffersizeperstream  = buffersizeperstream ;
                        msg_store_v2->nbstream             = requestedstreamnumber;
                        msg_store_v2->filesize             = filesize ;
#endif
                        if ( writemessage(outcontrolsock,receive_buffer,STORMESSLEN_V2,SENDCONTROLTO) < 0 ) {
                            syslog(LOG_ERR,"Error wtiting MSG_TRANS_OK_V2 second part") ;
                            storeunlink(realfilename) ;
                            free_all_var() ;
                            free(logmessage) ;
                            free(receive_buffer) ;
                            return -1 ;
                        }
                        if ( filesize == 0 ) {
                            state = S_WAITING_CREATE_ZERO ;
                            free_all_var() ;
                        } else {
                            state = S_WAITING_RETR_START ;
                        }
                        free(logmessage) ;
                        free(receive_buffer) ;
                        return 0 ;
                    }
                }
                default :{
                    syslog(LOG_ERR,"Unkown message in S_WAITING_FILENAME state : %d",msgcode) ;
                    return -1 ;
                }
            }
        }
        
        /*
        ** Case S_WAITING_STORE_START
        **
        **         MSG_TRANS_START_V2
        **         MSG_ABORT
        */
        case S_WAITING_STORE_START : {
             switch (msgcode) {
                case MSG_TRANS_START_V2 :{
                    syslog(LOG_NOTICE,"Receiving MSG_TRANS_START_V2") ;
                    if ( msglen != requestedstreamnumber*sizeof(int)) {
                        storeunlink(realfilename) ;
                        free_all_var() ;
                        reply(MSG_BAD,"Inconsistency between MSG_TRANS_START_V2 and message length") ;
                        syslog(LOG_ERR,"Inconsistency between MSG_TRANS_START_V2 and message length") ;
                        return -1 ;
                    }
                    /*
                    ** Alloc space for logmessage 
                    */
                    if ( (logmessage = (char *) malloc ( 1024 )) == NULL ) {
                        storeunlink(realfilename) ;
                        free_all_var() ;
                        reply(MSG_BAD,"Error allocating memory for logmessage") ;
                        syslog(LOG_ERR,"Error allocating memory for logmessage") ;
                        return -1 ;
                    }
                    /*
                    ** Retreive port numbers
                    */
                     if ( readmessage(incontrolsock,(char *) myports,msglen,CONTROLSOCKTO) < 0 ) {
                        storeunlink(realfilename) ;
                        free_all_var() ;
                        free(logmessage) ;
                        syslog(LOG_ERR,"Error reading MSG_TRANS_START_V2") ;
                        return -1 ;
                    } else {
#ifndef WORDS_BIGENDIAN
                        portfree = myports ;
                        for (i=0 ; i < requestedstreamnumber ; i++ ) {
                            *portfree = ntohl(*portfree) ;
                            portfree++ ;
                        }
#endif
                        portfree = myports ;
                        for (i=0 ; i < requestedstreamnumber ; i++ ) {
                            portfree++ ;
                        }
                    }
                    if ( (retcode = storetransferfile(realfilename,logmessage)) == 0 ) {
                        /*
                        ** Everything seems to work correctly
                        */
                        free(logmessage) ;
                        return 0 ;
                    }
                    if ( retcode > 0 ) {
                        /*
                        ** Something goes wrong
                        */
                        storeunlink(realfilename) ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    }
                    /*
                    ** 
                    */
                    storeunlink(realfilename) ;
                    free_all_var() ;
                    free(logmessage) ;
                    state = S_LOGGED ;
                    return -1 ;
                }
                case MSG_ABORT : {
                    syslog(LOG_ERR,"Receive ABORT message") ;
                    storeunlink(realfilename) ;
                    free_all_var() ;
                    state = S_LOGGED ;
                    return 0 ;
                }
                default :{
                    syslog(LOG_ERR,"Unkown message in S_WAITING_STORE_START state : %d",msgcode) ;
                    storeunlink(realfilename) ;
                    free_all_var() ;
                    return -1 ;
                }
            }
        }
        /*
        ** Case S_WAITING_CREATE_ZERO
        **
        **         MSG_CREATE_ZERO
        **         MSG_ABORT
        */
        case S_WAITING_CREATE_ZERO : {
             switch (msgcode) {
                case MSG_CREATE_ZERO : {
                    syslog(LOG_ERR,"Receive MSG_CREATE_ZERO message ") ;
                    state = S_LOGGED ;
                    return 0 ;
                }
                case MSG_ABORT : {
                    syslog(LOG_ERR,"Receive ABORT message") ;
                    state = S_LOGGED ;
                    return 0 ;
                }
                default :{
                    syslog(LOG_ERR,"Unkown message in S_WAITING_CREATE_ZERO state : %d",msgcode) ;
                    return -1 ;
                }
            }
        }
        /*
        ** Case S_WAITING_RETR_START
        **
        **         MSG_TRANS_START_V2
        **         MSG_ABORT
        */
        case S_WAITING_RETR_START : {
             switch (msgcode) {
                case MSG_TRANS_START_V2 : {
                    syslog(LOG_NOTICE,"Receiving MSG_TRANS_START_V2") ;
                    if ( msglen != requestedstreamnumber*sizeof(int)) {
                        free_all_var() ;
                        reply(MSG_BAD,"Inconsistency between MSG_TRANS_START_V2 and message length") ;
                        syslog(LOG_ERR,"Inconsistency between MSG_TRANS_START_V2 and message length") ;
                        return -1 ;
                    }
                    /*
                    ** Alloc space for logmessage 
                    */
                    if ( (logmessage = (char *) malloc ( 1024 )) == NULL ) {
                        free_all_var() ;
                        reply(MSG_BAD,"Error allocating memory for logmessage") ;
                        syslog(LOG_ERR,"Error allocating memory for logmessage") ;
                        return -1 ;
                    }
                    /*
                    ** Retreive port numbers
                    */
                     if ( readmessage(incontrolsock,(char *) myports,msglen,CONTROLSOCKTO) < 0 ) {
                        free_all_var() ;
                        free(logmessage) ;
                        syslog(LOG_ERR,"Error reading MSG_TRANS_START_V2") ;
                        return -1 ;
                    } else {
#ifndef WORDS_BIGENDIAN
                        portfree = myports ;
                        for (i=0 ; i < requestedstreamnumber ; i++ ) {
                            *portfree = ntohl(*portfree) ;
                            portfree++ ;
                        }
#endif
                        portfree = myports ;
                        for (i=0 ; i < requestedstreamnumber ; i++ ) {
                            portfree++ ;
                        }
                    }
                    if ( (retcode = retrtransferfile(curfilename,logmessage)) == 0 ) {
                        /*
                        ** Everything seems to work correctly
                        */
                        syslog(LOG_ERR,"retrtransferfile retcode = 0") ;
                       
                        free(logmessage) ;
                        return 0 ;
                    }
                    if ( retcode > 0 ) {
                        /*
                        ** Something goes wrong
                        */
                        syslog(LOG_ERR,"retrtransferfile retcode > 0") ;
                        free_all_var() ;
                        free(logmessage) ;
                        state = S_LOGGED ;
                        return 0 ;
                    }
                    /*
                    ** 
                    */
                    syslog(LOG_ERR,"retrtransferfile retcode < 0") ;
                    free_all_var() ;
                    free(logmessage) ;
                    state = S_LOGGED ;
                    return -1 ;
                }
                case MSG_ABORT : {
                    syslog(LOG_ERR,"Receive ABORT message") ;
                    free_all_var() ;
                    state = S_LOGGED ;
                    return 0 ;
                }
                default :{
                    syslog(LOG_ERR,"Unkown message in S_WAITING_RETR_START state : %d",msgcode) ;
                    return -1 ;
                }
            }
        }
        /*
        ** Case S_SENDING or S_RECEIVING
        */
        case S_SENDING :
        case S_RECEIVING : {
              switch (msgcode) {
                case MSG_ABORT :{
                    syslog(LOG_ERR,"Receive MSG_ABORT while sending or receiving") ;
                    return -1 ;
                }
                default :{
                    syslog(LOG_ERR,"Unkown message in S_SENDING or S_RECEIVING state : %d",msgcode) ;
                    return -1 ;
                }    
            }
        }
        /*
        ** Any other state
        */
        default : {
            syslog(LOG_ERR,"Receive message in state %d",state) ;
            return -1 ;
        }
    }
}
       
        
