/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 readmessage.c v 0.0.0  1999/11/24
               v 1.6.0  2000/03/24  - Take the readmessage routine from 
                                      bbftpc which is more flexible the
                                      only difference is that this one is
                                      logging into syslog
               v 1.6.1  2000/03/24  - Portage to OSF1
                                    - Set the socket in non block mode
               v 1.8.7  2000/05/30  - Portage to IRIX64
               v 1.8.10 2000/08/11  - Portage to Linux
               v 1.9.0  2000/08/24  - Use configure to help port
               v 1.9.4  2000/10/16  - Supress %m
               v 2.0.0  2000/12/07  - Introduce discardmessage routine
                                    - Replace recv by read for ssh mode
                                      (Tim Adye)
               v 2.0.1 2001/04/23   - Correct indentation

*****************************************************************************/
#include <errno.h>
#include <syslog.h>
#include <sys/time.h>
#include <unistd.h>

#include <daemon.h>


int readmessage(int sock,char *buffer,int msglen,int to) 
{
    int    nbget ;
    int    retcode ;
    int        nfds ;
    fd_set    selectmask ; /* Select mask */
    struct timeval    wait_timer;

    nbget = 0 ;
/*
** start the reading loop
*/
    while ( nbget < msglen ) {
        nfds = sysconf(_SC_OPEN_MAX) ;
        FD_ZERO(&selectmask) ;
        FD_SET(sock,&selectmask) ;
        wait_timer.tv_sec  = to ;
        wait_timer.tv_usec = 0 ;
        if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
            /*
            ** Select error
            */
            syslog(LOG_ERR,"Read message : Select error : MSG (%d,%d)",msglen,nbget) ;
            return -1 ;
        } else if ( retcode == 0 ) {
            syslog(LOG_ERR,"Read message : Time out : MSG (%d,%d)",msglen,nbget) ;
            return -1 ;
        } else {
            retcode = read(sock,&buffer[nbget],msglen-nbget) ;
            if ( retcode < 0 ) {
                syslog(LOG_ERR,"Read message : Receive error : MSG (%d,%d) : %s",msglen,nbget,strerror(errno)) ;
                return -1 ;
            } else if ( retcode == 0 ) {
                syslog(LOG_ERR,"Read message : Connexion breaks") ;
                return -1 ;
            } else {
                nbget = nbget + retcode ;
            }
        }
    }
    return(0) ;
}

int discardmessage(int sock,int msglen,int to) 
{
    int    nbget ;
    int    retcode ;
    int        nfds ;
    fd_set    selectmask ; /* Select mask */
    struct timeval    wait_timer;
    char buffer[256] ;
    
    /*
    ** We are going to read buflen by buflen till the message 
    ** is exhausted
    */
    
    nbget = 0 ;
    while ( nbget < msglen ) {
        nfds = sysconf(_SC_OPEN_MAX) ;
        FD_ZERO(&selectmask) ;
        FD_SET(sock,&selectmask) ;
        wait_timer.tv_sec  = to ;
        wait_timer.tv_usec = 0 ;
        if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
            /*
            ** Select error
            */
            syslog(LOG_ERR,"Discard message : Select error : MSG (%d,%d)",msglen,nbget) ;
            return -1 ;
        } else if ( retcode == 0 ) {
            syslog(LOG_ERR,"Discard message : Time out : MSG (%d,%d)",msglen,nbget) ;
            return -1 ;
        } else {
            retcode = recv(sock,buffer,sizeof(buffer),0) ;
            if ( retcode < 0 ) {
                syslog(LOG_ERR,"Discard message : Receive error : MSG (%d,%d) : %s",msglen,nbget,strerror(errno)) ;
                return -1 ;
            } else if ( retcode == 0 ) {
                syslog(LOG_ERR,"Discard message : Connexion breaks") ;
                return -1 ;
            } else {
                nbget = nbget + retcode ;
            }
        }
    }
    return(0) ;
}
        
    
