/****************************************************************************

 Copyright (c) 1999 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 sendcrypt.c v 1.4.0  2000/03/22
             v 1.6.1  2000/03/28    - Portage to OSF1
             v 1.8.0  2000/04/14    - Introduce RSA Cryptage
             v 1.8.2  2000/04/17    - Portage to OSF1
             v 1.8.4  2000/04/21    - Random seed done in do_deamon.c
             v 1.8.7  2000/05/24    - Modify headers
             v 1.8.10 2000/08/11    - Portage to Linux
             v 1.9.0  2000/08/18    - Use configure to help portage
             v 2.0.0  2000/12/18    - Use incontrolsock and outcontrolsock
             v 2.0.1  2001/04/23    - Correct indentation

*****************************************************************************/
#include <errno.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <syslog.h>
#include <sys/time.h>

#include <bbftpd.h>
#include <common.h>
#include <daemon.h>
#include <structures.h>
#include <openssl/rsa.h>
#include <openssl/err.h>
#include <openssl/rand.h>

extern int  outcontrolsock ;
extern RSA    *myrsa ;

void sendcrypt() 
{
    struct message    *mess ;
    struct mess_sec    *msg_sec ;
    char    buf[MAXMESSLEN] ;
    unsigned char    pubkey[NBITSINKEY] ;
    unsigned char    pubexponent[NBITSINKEY] ;
    int        lenkey ;
    int        lenexpo ;
    
    /*
    ** Ask for the private and public Key
    */
    if ( (myrsa = RSA_generate_key(NBITSINKEY,3,NULL,NULL)) == NULL) {
        syslog(LOG_ERR,"%s",ERR_error_string(ERR_get_error(),NULL) ) ;
        exit(1) ;
    }
    /*
    ** Now extract the public key in order to send it
    */
    lenkey  = BN_bn2mpi(myrsa->n,pubkey) ;
    lenexpo = BN_bn2mpi(myrsa->e,pubexponent) ;
    mess = (struct message *) buf ;
    mess->code = MSG_CRYPT ;
#ifndef WORDS_BIGENDIAN
    mess->msglen = ntohl(CRYPTMESSLEN+lenkey+lenexpo) ;
#else
    mess->msglen = CRYPTMESSLEN+lenkey+lenexpo ;
#endif
    if (writemessage(outcontrolsock,buf,MINMESSLEN,SENDCONTROLTO) < 0 ) {
        syslog(LOG_ERR,"Error on sendcrypt 1") ;
        exit(1) ;
    }
    msg_sec  = (struct mess_sec    *) buf ;
    msg_sec->crtype  = CRYPT_RSA_PKCS1_OAEP_PADDING ;
#ifndef WORDS_BIGENDIAN
    msg_sec->pubkeylen  = ntohl(lenkey) ;
    msg_sec->expolen  = ntohl(lenexpo) ;
#else
    msg_sec->pubkeylen  = lenkey ;
    msg_sec->expolen  = lenexpo ;
#endif
    if (writemessage(outcontrolsock,buf,CRYPTMESSLEN,SENDCONTROLTO) < 0 ) {
        syslog(LOG_ERR,"Error on sendcrypt 2") ;
        exit(1) ;
    }
    /*
    ** Send Key and exponent
    */
    if (writemessage(outcontrolsock,pubkey,lenkey,SENDCONTROLTO) < 0 ) {
        syslog(LOG_ERR,"Error on sendcrypt pubkey") ;
        exit(1) ;
    }
    if (writemessage(outcontrolsock,pubexponent,lenexpo,SENDCONTROLTO) < 0 ) {
        syslog(LOG_ERR,"Error on sendcrypt pubexponent") ;
        exit(1) ;
    }
}
