/****************************************************************************

 Copyright (c) 1999,2000 Gilles Farrache In2p3 Computing Center
 All rights reserved.
  
  
 signals_routines_v2.c v 2.0.0 2000/12/14  - Replace files set_signals.c and
                                             set_father_signals.c
                       v 2.0.1 2001/04/23  - Correct indentation

 
*****************************************************************************/
#include <errno.h>
#include <signal.h>
#include <sys/wait.h>
#include <syslog.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <utime.h>

#include <bbftpd.h>
#include <config.h>
#include <daemon.h>
#include <status.h>
#include <structures.h>

#ifdef RFIO
#include <shift.h>
#endif

/*
** For V1 and V2 Protocol
*/
extern int  flagsighup;
extern int  childendinerror ;
extern int  state ;
extern int  killdone ;
extern int  unlinkfile ;
extern pid_t    fatherpid ;
/*
** For V2 protocol
*/
extern char *realfilename ;
extern char *curfilename ;
extern int  transferoption ;
extern int  *mychildren ;
extern int  nbpidchild ;
extern char lastaccess[9] ;
extern char lastmodif[9] ;
extern int  filemode ;

void clean_child() ; 
int checkendchild(int status) ;

void in_sigchld_v2(sig) 
int        sig ;
{
    int *pidfree ;
    int    pid ;
    int    status ;
    int    totpid ;
    int i ;
    char logmessage[1024] ;
    int    retcode ;
    struct utimbuf ftime ;
    int goon ;
    /*
    ** We are going to check if the process generating this
    ** signal was not already detected as dead. This was causing 
    ** a send of a MSG_OK if all process have been dectected dead
    ** througn the waitpid routine and the signal was pending.
    ** If all processes have been detected dead before childendinerror 
    ** to 0, totpid also so we send a reply MSS_OK
    */
    totpid = 0 ;
    pidfree = mychildren ;
    for ( i=0 ; i< nbpidchild ; i++) {
        totpid = totpid + *pidfree ;
        pidfree++ ;
    }
    if ( totpid == 0 ) {
        /*
        ** All Ended
        */
        free_all_var() ;
        return ;
    }
/*
** Check the status
*/
    pidfree = mychildren ;
    for ( i=0 ; i< nbpidchild; i++) {
        if ( *pidfree != 0 ) {
            pid = waitpid(*pidfree,&status,WNOHANG) ;
            if ( pid == *pidfree ) {
                /*
                ** the process has ended
                */
                if ( (retcode = checkendchild(status)) != 0 ) {
                    if ( childendinerror == 0 ) {
                        childendinerror = 1 ;
                        syslog(LOG_ERR,"Child pid %d ends in error status %d",pid,retcode) ;
                        if ( unlinkfile != 0 ) storeunlink(realfilename) ;
                        if (retcode == 255) {
                            sprintf(logmessage,"Disk quota excedeed or No Space left on device") ;
                            reply(MSG_BAD_NO_RETRY,logmessage) ;
                        } else {
                            if (retcode ==  ENOEXEC ) {
                                sprintf(logmessage,"Interrupted by signal") ;
                            } else {
                                sprintf(logmessage,"%s",strerror(retcode)) ;
                            }
                            reply(MSG_BAD,logmessage) ;
                        }
                    } else {
                        syslog(LOG_ERR,"Child pid %d ends in error",pid) ;
                    }
                }
                *pidfree = 0 ;
            }
        }
        pidfree++ ;
    }
    totpid = 0 ;
    pidfree = mychildren ;
    for ( i=0 ; i< nbpidchild ; i++) {
        totpid = totpid + *pidfree ;
        pidfree++ ;
    }
    if ( totpid == 0 && childendinerror == 0 ) {
        if ((unlinkfile == 1 || unlinkfile == 2)) {
            goon = 0 ;
            if ( (transferoption & TROPT_ACC ) == TROPT_ACC ) {
                sscanf(lastaccess,"%08x",&ftime.actime) ;
                sscanf(lastmodif,"%08x",&ftime.modtime) ;
                if ( storeutime(realfilename,&ftime,logmessage) < 0 ) {
                    storeunlink(realfilename) ;
                    reply(MSG_BAD,logmessage) ;
                    goon = 1 ;
                }
           }
           if ( (goon == 0 ) && ((transferoption & TROPT_MODE ) == TROPT_MODE) ) {
                if ( storechmod(realfilename,filemode,logmessage) < 0 ) {
                    storeunlink(realfilename) ;
                    reply(MSG_BAD,logmessage) ;
                    goon = 1 ;
                }
            }
            if ( (goon == 0 ) && ((transferoption & TROPT_TMP ) == TROPT_TMP ) ) {
                if ( storerename(realfilename,curfilename,logmessage) < 0 ) {
                    storeunlink(realfilename) ;
                    reply(MSG_BAD,logmessage) ;
                    goon = 1 ;
                }
            }
            state = S_LOGGED ;
            if ( goon == 0 ) reply(MSG_OK,"OK") ;
            free_all_var() ;
        } else {
            state = S_LOGGED ;
            reply(MSG_OK,"OK") ;
            free_all_var() ;
        }
    } 
    if ( childendinerror == 1 ) {
        clean_child() ;
    }
    if (totpid == 0 ) {
        state = S_LOGGED ;
        unlinkfile = 0 ;
        killdone = 0 ;
        childendinerror = 0 ;
        free_all_var() ;
    }
}

void in_sighup_v2(sig) 
int        sig ;
{
    /* Receive sighup in order to restart*/
    /*
    syslog(LOG_NOTICE,"Receive SIGHUP") ;
    */
    flagsighup = 1 ;
}

void in_sigterm_v2(sig) 
int        sig ;
{
    if ( fatherpid == getpid() ) {
        /*
        ** We are in father
        */
        /*
        ** Unlink the file if we are getting a file
        */ 
        if ( unlinkfile == 1 ) unlink(realfilename) ;
        clean_child() ;
        exit() ;
    } else {
        exit(EINTR) ;
    }
}

int set_signals_v2() {
    struct    sigaction    sga ;
    
    sga.sa_handler = SIG_IGN ;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0   ;
    if ( sigaction(SIGPIPE,&sga,0) < 0 ) {
        syslog(LOG_ERR,"Error sigaction SIGPIPE : %s",strerror(errno)) ;
        return(-1) ;
    }
    sga.sa_handler = in_sigchld_v2 ;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0  ;
    if ( sigaction(SIGCHLD,&sga,0) < 0 ) {
        syslog(LOG_ERR,"Error sigaction SIGCHLD : %s",strerror(errno)) ;
        return(-1) ;
    }
    sga.sa_handler = in_sighup_v2 ;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0  ;
    if ( sigaction(SIGHUP,&sga,0) < 0 ) {
        syslog(LOG_ERR,"Error sigaction SIGHUP : %s",strerror(errno)) ;
        return(-1) ;
    }
    sga.sa_handler = in_sigterm_v2 ;
    sigemptyset(&(sga.sa_mask));
    sga.sa_flags = 0  ;
    if ( sigaction(SIGTERM,&sga,0) < 0 ) {
        syslog(LOG_ERR,"Error sigaction SIGTERM : %s",strerror(errno)) ;
        return(-1) ;
    }
    return 0 ;
}
