/****************************************************************************

 Copyright (c) 1999,2000,2001 Gilles Farrache In2p3 Computing Center
 All rights reserved.

 Routines : int storecheckoptions_rfio(char *logmessage)  
            int storemkdir_rfio(char *dirname,char *logmessage,int recursif) 
            int storecreatefile_rfio(char *filename, char *logmessage) 
            int storetransferfile_rfio(char *filename, char *logmessage) 

 store_rfio_v2.c v 2.0.0 2000/12/19 - Routines creation
                 v 2.0.1 2001/04/17 - Realy wait STARTCHILDTO 
                                    - Correct indentation
                                    - Port to IRIX
               
*****************************************************************************/
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <signal.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <bbftpd.h>
#include <common.h>
#include <daemon.h>
#include <status.h>
#include <config.h> 
#include <structures.h>

#ifdef RFIO
#include <shift.h>
#endif

extern int  transferoption ;
extern my64_t  filesize ;
extern int  requestedstreamnumber ;
extern int  buffersizeperstream ;
extern int  maxstreams ;
extern int  *myports ;
extern char *readbuffer ;
extern char *compbuffer ;
extern int  *mychildren ;
extern int  nbpidchild ;
extern int  unlinkfile ;
extern int  incontrolsock ;
extern int  outcontrolsock ;
extern int  state ;
extern int childendinerror ;
extern int flagsighup ;
extern int mycos ;

#include <zlib/zlib.h>

/*******************************************************************************
** storechmod_rfio :                                                           *
**                                                                             *
**      Routine to chmod a file                                                *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int storechmod_rfio(char *filename,int mode,char *logmessage)
{
#ifdef RFIO
    int     retcode ;
    
    retcode = rfio_chmod(filename,mode) ;
    if ( retcode < 0 ) {
        sprintf(logmessage,"Error chmod on file %s : %s",filename,strerror(errno)) ;
        syslog(LOG_ERR,"Error chmod on file %s : %s",filename,strerror(errno)) ;
        return retcode ;
    }
#endif
    return 0 ;
}

/*******************************************************************************
** storerename :                                                               *
**                                                                             *
**      Routine to unlink a file                                               *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int storerename_rfio(char *newfilename,char *oldfilename,char *logmessage)
{
    
#ifdef RFIO
    int     retcode ;
    
    retcode = rfio_rename(newfilename,oldfilename) ;
    if ( retcode < 0 ) {
        sprintf(logmessage,"Error renaming %s to %s : %s",newfilename,oldfilename,strerror(errno)) ;
        syslog(LOG_ERR,"Error renaming %s to %s : %s",newfilename,oldfilename,strerror(errno)) ;
        return retcode ;
    }
#endif
    return 0 ;
}
/*******************************************************************************
** storeunlink_rfio :                                                          *
**                                                                             *
**      Routine to unlink a file                                               *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**                                                                             *
**      RETURN:                                                                *
**          -1  Failed                                                         *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int storeunlink_rfio(char *filename,char *logmessage)
{
#ifdef RFIO
    rfio_unlink(filename) ;
#endif
    return 0 ;
}
/*******************************************************************************
** storecheckoptions_rfio :                                                    *
**                                                                             *
**      Routine to check the options parameters                                *
**                                                                             *
**      RETURN:                                                                *
**          -1  Incorrect options                                              *
**           0  OK                                                             *
**                                                                             *
*******************************************************************************/

int storecheckoptions_rfio(char *logmessage)
{
#ifndef RFIO
    /*
    ** Just reply that RFIO is not supported
    */
    syslog(LOG_NOTICE,"Fail to store : RFIO not supported") ;
    sprintf(logmessage,"Fail to store : RFIO not supported") ;
    return -1 ;
#else
    int tmpoptions ;
    
    /*
    ** Check the compatibility with other options
    */
    if ((transferoption & TROPT_ACC) == TROPT_ACC) {
        /*
        ** RFIO has no fonction to change access and modif time
        ** till now to have the same behaviour in deamon and client
        ** do not care
        */
    }
   /*
    ** Check all known options
    */
    tmpoptions = TROPT_TMP | TROPT_ACC | TROPT_MODE | TROPT_DIR | TROPT_GZIP | TROPT_RFIO ;
    transferoption = transferoption & tmpoptions ;
    /*
    ** Check the maxstreams
    */
    if ( requestedstreamnumber > maxstreams ) requestedstreamnumber = maxstreams ;
    return 0 ;
#endif
}       
/*******************************************************************************
** storemkdir_rfio :                                                                *
**                                                                             *
**      Routine to create a bunch of directory                                 *
**                                                                             *
**      INPUT variable :                                                       *
**          dirname    :  directory to create    NOT MODIFIED                  *
**          recursif   :  1 recursif             NOT MODIFIED                  *
**                        0 non recursif                                       *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      TO FREE before any return : dirpath                                    *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**           >0  Creation failed recoverable error                             *
**                                                                             *
*******************************************************************************/
int storemkdir_rfio(char *dirname,char *logmessage,int recursif)
{
#ifndef RFIO
    /*
    ** Just reply that RFIO is not supported
    */
    syslog(LOG_NOTICE,"Fail to mkdir : RFIO not supported") ;
    sprintf(logmessage,"Fail to mkdir : RFIO not supported") ;
    return -1 ;
#else
    char    *dirpath;
    char    *basedir ;
    char    *slash ;
    int     savederrno ;
    struct stat statbuf;
    int     k ;
    
    /*
    ** make a copy of dirname for security
    */
    if ( (dirpath = (char *) malloc (strlen(dirname)+1) ) == NULL ) {
        sprintf(logmessage,"Error allocating memory for dirpath : %s",strerror(errno)) ;
        return 1 ;
    }
    strcpy (dirpath, dirname);
    /*
    ** Strip trailing slash
    */
    strip_trailing_slashes(dirpath) ;
    slash = dirpath ;
    if ( recursif == 1 ) {
        /*
        ** Recursif, we are going to create all directory
        */
        /*
        ** This is an absolut path, but look if it is a syntax like :
        **      host:/.....
        **      /.....
        */
        if ( *dirpath == '/' ) {
            while (*slash == '/') slash++;
        } else {
            /*
            ** find the :
            */
            k = 0 ;
            while (*slash != ':' && k < strlen(dirpath) ) {
                slash++;
                k++ ;
            }
            if ( k == strlen(dirpath) ) {
                sprintf(logmessage,"Incorrect directory name : %s",dirpath) ;
                free(dirpath) ;
                return -1 ;
            } else {
                slash++ ;
                slash++ ;
            }
        }
        while (1) {
            slash = (char *) strchr (slash, '/');
            if ( slash == NULL ) break ;
            basedir = dirpath ;
            /*
            ** The mkdir and stat calls below appear to be reversed.
            ** They are not.  It is important to call mkdir first and then to
            ** call stat (to distinguish the three cases) only if mkdir fails.
            ** The alternative to this approach is to `stat' each directory,
            ** then to call mkdir if it doesn't exist.  But if some other process
            ** were to create the directory between the stat & mkdir, the mkdir
            ** would fail with EEXIST. 
            ** We create it with mode = 777 because the umask is set by bbftpd
            */
             *slash = '\0';
            if ( rfio_mkdir (basedir, 0777) ) {
                savederrno = errno ;
                if ( rfio_stat (basedir,&statbuf ) ) {
                    sprintf(logmessage,"Error creation directory %s (%s)",basedir,strerror(savederrno)) ; 
                    /*
                    ** We tell the client not to retry in the following case (even in waiting
                    ** WAITRETRYTIME the problem will not be solved) :
                    **        EACCES        : Search permission denied
                    **        EDQUOT        : No more quota 
                    **        ENOSPC        : No more space
                    **        ELOOP        : To many symbolic links on path
                    **        ENAMETOOLONG: Path argument too long
                    **        ENOTDIR        : A component in path is not a directory
                    **        EROFS        : The path prefix resides on a read-only file system.
                    **        ENOENT      : A component of the path prefix does not exist or is a null pathname.
                    **        EEXIST      : The named file already exists.
                    */
                     if ( savederrno == EACCES ||
                            savederrno == EDQUOT ||
                            savederrno == ENOSPC ||
                            savederrno == ELOOP ||
                            savederrno == ENAMETOOLONG ||
                            savederrno == ENOTDIR ||
                            savederrno == EROFS ||
                            savederrno == ENOENT ||
                            savederrno == EEXIST ) {
                        free(dirpath) ;
                        return -1 ;
                    } else {
                        free(dirpath) ;
                        return 1 ;
                    }
                } else if ( (statbuf.st_mode & S_IFDIR) != S_IFDIR) {
                    sprintf(logmessage,"%s exists but is not a directory",basedir) ; 
                    free(dirpath) ;
                    return -1 ;
                } else {
                     /*
                    ** dirpath already exists and is a directory. 
                    */
                }
            }
             *slash++ = '/';

            /* 
            ** Avoid unnecessary calls to `stat' when given
            ** pathnames containing multiple adjacent slashes.  
            */
            while (*slash == '/') slash++;
       }
    }
    /*
    ** We have created all leading directories so let see the last one
    */
    basedir = dirpath ;
    if ( rfio_mkdir (basedir, 0777) ) {
        savederrno = errno ;
        if ( rfio_stat (basedir,&statbuf ) ) {
            sprintf(logmessage,"Error creation directory %s (%s)",basedir,strerror(savederrno)) ; 
            /*
            ** We tell the client not to retry in the following case (even in waiting
            ** WAITRETRYTIME the problem will not be solved) :
            **        EACCES        : Search permission denied
            **        EDQUOT        : No more quota 
            **        ENOSPC        : No more space
            **        ELOOP        : To many symbolic links on path
            **        ENAMETOOLONG: Path argument too long
            **        ENOTDIR        : A component in path is not a directory
            **        EROFS        : The path prefix resides on a read-only file system.
            **        ENOENT      : A component of the path prefix does not exist or is a null pathname.
            **        EEXIST      : The named file already exists.
            */
             if ( savederrno == EACCES ||
                    savederrno == EDQUOT ||
                    savederrno == ENOSPC ||
                    savederrno == ELOOP ||
                    savederrno == ENAMETOOLONG ||
                    savederrno == ENOTDIR ||
                    savederrno == EROFS ||
                    savederrno == ENOENT ||
                    savederrno == EEXIST ) {
                free(dirpath) ;
                return -1 ;
            } else {
                free(dirpath) ;
                return 1 ;
            }
        } else if ( (statbuf.st_mode & S_IFDIR) != S_IFDIR) {
            sprintf(logmessage,"%s exists but is not a directory",basedir) ; 
            free(dirpath) ;
            return -1 ;
        } else {
             /*
            ** dirpath already exists and is a directory. 
            */
            free(dirpath) ;
            return 0 ;
        }
    }
    free(dirpath) ;
    return 0 ;
#endif
}
/*******************************************************************************
** storecreatefile_rfio :                                                      *
**                                                                             *
**      Routine to create a file                                               *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**          transferoption                   NOT MODIFIED                      * 
**          filesize                         NOT MODIFIED                      *
**                                                                             *
**      TO FREE before any return : filepath                                   *
**                                                                             *
**      RETURN:                                                                *
**          -1  Transfer failed calling has to close the connection            *
**           0  OK  (Does not mean the transfer is successfull                 *
**                                                                             *
*******************************************************************************/

int storecreatefile_rfio(char *filename, char *logmessage) 
{
#ifndef RFIO
    /*
    ** Just reply that RFIO is not supported
    */
    syslog(LOG_NOTICE,"Fail to create file : RFIO not supported") ;
    sprintf(logmessage,"Fail to create file : RFIO not supported") ;
    return -1 ;
#else
    char    *filepath ;
    int     fd ;
    int     lastslash ;
    int     savederrno ;
    int     retcode ;
    off_t   toseek ;
    struct stat statbuf ;
    
    /*
    ** make a copy of filename for security
    */
    if ( (filepath = (char *) malloc (strlen(filename)+1) ) == NULL ) {
        sprintf(logmessage,"Error allocating memory for filepath : %s",strerror(errno)) ;
        return 1 ;
    }
    strcpy (filepath, filename);
    /*
    ** Check if the file exist
    **      We are going to first see if the directory exist
    **      and then look at the options 
    **      and then look if the file exist
    */
    lastslash = strlen(filepath) - 1 ;
    while ( lastslash >= 0 && filepath[lastslash] != '/') lastslash-- ;
    if ( lastslash == -1 ) {
        /*
        ** No slash in the path, that means that we have done a chdir
        ** before so do not care for the directory problem
        */
    } else if ( lastslash == 0 ) {
        /*
        ** A slash in first position so we suppose the "/" directory 
        ** exist ... nothing to do
        */
    } else if ( lastslash == strlen(filepath) - 1 ) {
        /*
        ** The filename end with a slash ..... error
        */
        syslog(LOG_ERR,"Filename %s ends with a /",filepath) ;
        sprintf(logmessage,"Filename %s ends with a /",filepath) ;
        free(filepath) ;
        return -1 ;
    } else {
        filepath[lastslash] = '\0';
        /*
        ** Check the existence of the directory 
        */
        if ( rfio_stat(filepath,&statbuf) < 0 ) {
            /*
            ** It may be normal to get an error if the directory
            ** does not exist but some error code must lead
            ** to the interruption of the transfer:
            **        EACCES        : Search permission denied
            **        ELOOP        : To many symbolic links on path
            **        ENAMETOOLONG: Path argument too long
            **        ENOTDIR        : A component in path is not a directory
            */
             savederrno = errno ;
            if ( savederrno == EACCES ||
                savederrno == ELOOP ||
                savederrno == ENAMETOOLONG ||
                savederrno == ENOTDIR ) {
                syslog(LOG_ERR,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
                sprintf(logmessage,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
                free(filepath) ;
                return -1 ;
            } else if (savederrno == ENOENT) {
                /*
                ** The directory does not exist so check for the TROPT_DIR
                ** option 
                */
                if ( (transferoption & TROPT_DIR ) != TROPT_DIR ) {
                    syslog(LOG_ERR,"Directory (%s) creation needed but TROPT_DIR not set",filepath) ;
                    sprintf(logmessage,"Directory (%s) creation needed but TROPT_DIR not set",filepath) ;
                    free(filepath) ;
                    return -1 ;
                } else {
                    if ( (retcode = storemkdir_rfio(filepath,logmessage,1)) != 0 ) {
                           syslog(LOG_ERR,"%s",logmessage) ;
                        free(filepath) ;
                        return retcode ;
                    }
                    filepath[lastslash] = '/';
                }
            } else {
                syslog(LOG_ERR,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
                sprintf(logmessage,"Error stating directory %s : %s ",filepath,strerror(savederrno)) ;
                free(filepath) ;
                return 1 ;
            }
        } else {
            /*
            ** The directory exist, check if it is a directory
            */
            if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
                /*
                ** OK correct
                */
                filepath[lastslash] = '/';
            } else {
                syslog(LOG_ERR,"%s is a not a directory ",filepath) ;
                sprintf(logmessage,"%s is a not a directory",filepath) ;
                free(filepath) ;
                return -1 ;
            }
        }
    }
    /*
    ** At this stage all directory exists so check for the file 
    */
    if ( rfio_stat(filepath,&statbuf) < 0 ) {
        /*
        ** It may be normal to get an error if the file
        ** does not exist but some error code must lead
        ** to the interruption of the transfer:
        **        EACCES        : Search permission denied
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG: Path argument too long
        **        ENOTDIR        : A component in path is not a directory
        */
        savederrno = errno ;
        if ( savederrno == EACCES ||
            savederrno == ELOOP ||
            savederrno == ENAMETOOLONG ||
            savederrno == ENOTDIR ) {
            sprintf(logmessage,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
            syslog(LOG_ERR,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
            free(filepath) ;
            return -1 ;
        } else if (savederrno == ENOENT) {
            /*
            ** That is normal the file does not exist
            */
        } else {
            sprintf(logmessage,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
            syslog(LOG_ERR,"Error stating file %s : %s ",filepath,strerror(savederrno)) ;
            free(filepath) ;
            return 1 ;
        }
    } else {
        /*
        ** The file exists so check if it is a directory
        */
        if ( (statbuf.st_mode & S_IFDIR) == S_IFDIR) {
            syslog(LOG_ERR,"file %s is a directory",filepath) ;
            sprintf(logmessage,"File %s is a directory",filepath) ;
            free(filepath) ;
            return -1 ;
        }
        /*
        ** In order to set correctly the user and group id we
        ** erase the file first
        */
        rfio_unlink(filepath) ;
    }                
    /*
    ** We create the file
    */
    if ((fd = rfio_open(filepath,O_WRONLY|O_CREAT|O_TRUNC,0777)) < 0 ) {
        /*
        ** Depending on errno we are going to tell the client to 
        ** retry or not
        */
        savederrno = errno ;
        sprintf(logmessage,"Error creation file %s : %s ",filepath,strerror(errno)) ;
        syslog(LOG_ERR,"Error creation file %s : %s ",filepath,strerror(errno)) ;
        /*
        ** We tell the client not to retry in the following case (even in waiting
        ** WAITRETRYTIME the problem will not be solved) :
        **        EACCES        : Search permission denied
        **        EDQUOT        : No more quota 
        **        ENOSPC        : No more space
        **        ELOOP        : To many symbolic links on path
        **        ENAMETOOLONG: Path argument too long
        **        ENOTDIR        : A component in path is not a directory
        */
        if ( savederrno == EACCES ||
                savederrno == EDQUOT ||
                savederrno == ENOSPC ||
                savederrno == ELOOP ||
                savederrno == ENAMETOOLONG ||
                savederrno == ENOTDIR ) {
                free(filepath) ;
                return -1 ;
        } else {
            free(filepath) ;
            return 1 ;
        }
    }
    if ( mycos > 0 ) {
        rfio_setcos(fd,0,mycos) ;
    }
    if ( filesize == 0 ) {
        rfio_close(fd) ;
        free(filepath) ;
        return 0 ;
    }
    /*
    ** Lseek to set it to the correct size
    ** We use toseek because filesize is of type my64t and 
    ** call to lseek can be done with off_t which may be
    ** of length 64 bits or 32 bits
    */
    toseek = filesize-1 ;
    if ( rfio_lseek(fd,toseek,SEEK_SET) < 0 ) {
        sprintf(logmessage,"Error seeking file %s : %s ",filepath,strerror(errno)) ;
        syslog(LOG_ERR,"Error seeking file %s : %s ",filepath,strerror(errno)) ;
        rfio_close(fd) ;
        rfio_unlink(filepath) ;
        free(filepath) ;
        return 1 ;
    }
    /*
    ** Write one byte 
    */
    if ( rfio_write(fd,"\0",1) != 1) {
        savederrno = errno ;
        sprintf(logmessage,"Error writing file %s : %s ",filepath,strerror(errno)) ;
        syslog(LOG_ERR,"Error writing file %s : %s ",filepath,strerror(errno)) ;
        rfio_close(fd) ;
        rfio_unlink(filepath) ;
        /*
        ** We tell the client not to retry in the following case (even in waiting
        ** WAITRETRYTIME the problem will not be solved) :
        **        EDQUOT        : No more quota 
        **        ENOSPC        : No space on device
        */
        if ( savederrno == EDQUOT ||
                savederrno == ENOSPC ) {
            free(filepath) ;
            return -1 ;
        } else {
            free(filepath) ;
            return 1 ;
        }
    }
    /*
    ** And close the file
    */
    if ( rfio_close(fd) < 0 ) {
        savederrno = errno ;
        rfio_unlink(filepath) ;
        sprintf(logmessage,"Error closing file %s : %s ",filepath,strerror(savederrno)) ;
        syslog(LOG_ERR,"Error closing file %s : %s ",filepath,strerror(savederrno)) ;
        if ( savederrno == ENOSPC ) {
            free(filepath) ;
            return -1 ;
        } else {
            free(filepath) ;
            return 1 ;
        }
    }
    return 0 ;
#endif
}
/*******************************************************************************
** storetransferfile_rfio :                                                    *
**                                                                             *
**      Routine to transfer a file                                             *
**                                                                             *
**      INPUT variable :                                                       *
**          filename    :  file to create    NOT MODIFIED                      *
**                                                                             *
**      OUTPUT variable :                                                      *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      GLOBAL VARIABLE USED :                                                 *                                                                      *
**                                                                             *
**      RETURN:                                                                *
**          -1  transfer failed unrecoverable error                            *
**           0  Keep the connection open (does not mean that the file has been *
**                successfully transfered)                                       *
**          >0  Recoverable error but calling has the cleaning to do           *
**                                                                             *
*******************************************************************************/
 
int storetransferfile_rfio(char *filename,char *logmessage) 
{
#ifdef RFIO
    off_t         nbperchild ;
    off_t         nbtoget;
    off_t         startpoint ;
    off_t         nbget ;
    off_t        toseek ;
    struct stat statbuf ;
    uLong    buflen ;
    uLong    bufcomplen ;
    int     lentowrite;
    int     lenwrited;
    int     datatoreceive;
    int     dataonone;

    int     *pidfree ;
    int     *portnumber ;
    int     i ;
    int     recsock ;
    int        retcode ;
    int        compressionon ;
    int     nfds ; 
    fd_set    selectmask ;
    struct timeval    wait_timer;
    int     fd ;
    my64_t    toprint64 ;
    struct mess_compress *msg_compress ;
    struct message *msg ;
    int     waitedtime ;

    portnumber = myports ;
    nbperchild = filesize/requestedstreamnumber ;
    pidfree = mychildren ;
    nbpidchild = 0 ;
    /*
    ** unlike is set to one in case of death of a children by a kill -9
    ** In this case the child is unable to unlink the file so that
    ** has to be done by the father
    */
    unlinkfile = 2 ;
    childendinerror = 0 ; /* No child so no error */
    for (i = 1 ; i <= requestedstreamnumber ; i++) {
        if ( i == requestedstreamnumber ) {
            startpoint = (i-1)*nbperchild;
            nbtoget = filesize-(nbperchild*(requestedstreamnumber-1)) ;
        } else {
            startpoint = (i-1)*nbperchild;
            nbtoget = nbperchild ;
        }
        /*
        ** Now create the socket to receive
        */
        recsock = 0 ;
        while (recsock == 0 ) {
            recsock = socketcreatereceive(*portnumber,logmessage) ;
        }
        if ( recsock < 0 ) {
            
            /*
            ** We set childendinerror to 1 in order to prevent the father
            ** to send a BAD message which can desynchronize the client and the
            ** server (We need only one error message)
            ** Bug discovered by amlutz on 2000/03/11
            */
            if ( childendinerror == 0 ) {
                childendinerror = 1 ;
                reply(MSG_BAD,logmessage) ;
            }
            clean_child() ;
            return 1 ;
        }
        portnumber++ ;
        /*
        ** Set flagsighup to zero in order to be able in child
        ** not to wait STARTCHILDTO if signal was sent before 
        ** entering select. (Seen on Linux with one child)
        */
        flagsighup = 0 ;
        if ( ( retcode = fork() ) == 0 ) {
            /*
            ** We are in child
            */
            /*
            ** Pause until father send a SIGHUP in order to prevent
            ** child to die before father has started all children
            */
            waitedtime = 0 ;
            while (flagsighup == 0 && waitedtime < STARTCHILDTO) {
                wait_timer.tv_sec  = 1 ;
                wait_timer.tv_usec = 0 ;
                nfds = sysconf(_SC_OPEN_MAX) ;
                select(nfds,0,0,0,&wait_timer) ;
                waitedtime = waitedtime + 1 ;
            }
            syslog(LOG_ERR,"Child %d starting",getpid()) ;
            /*
            ** Close all unnecessary stuff
            */
            close(incontrolsock) ;
            close(outcontrolsock) ; 
            /*
            ** Check if file exist
            */
            if ( rfio_stat(filename,&statbuf) < 0 ) {
                /*
                ** If the file does not exist that means that another
                ** child has detroyed it
                */ 
                i = errno ;
                syslog(LOG_ERR,"Error stating file %s : %s ",filename,strerror(errno)) ;
                close(recsock) ;
                exit(i) ;
            }
            /*
            ** Open and seek to position
            */
            if ((fd = rfio_open(filename,O_RDWR)) < 0 ) {
                i = errno ;
                rfio_unlink(filename) ;
                syslog(LOG_ERR,"Error opening file %s : %s ",filename,strerror(errno)) ;
                /*
                ** At this point a non recoverable error is 
                **        EDQUOT        : No more quota 
                **        ENOSPC        : No more space
                */
                if ( i == EDQUOT ||
                        i == ENOSPC ) {
                    close(recsock) ;
                    exit(255) ;
                } else {
                    close(recsock) ;
                    exit(i) ;
                }
            }
            if ( rfio_lseek(fd,startpoint,SEEK_SET) < 0 ) {
                i = errno ;
                rfio_close(fd) ;
                rfio_unlink(filename) ;
                syslog(LOG_ERR,"error seeking file : %s",strerror(errno)) ;
                close(recsock) ;
                exit(i)  ;
            }
            /*
            ** start the reading loop
            */
            nbget = 0 ;
            while ( nbget < nbtoget) {
                if ( (transferoption & TROPT_GZIP ) == TROPT_GZIP ) {
                    /*
                    ** Receive the header first 
                    */
                    if (readmessage(recsock,readbuffer,COMPMESSLEN,DATASOCKTO) < 0 ) {
                        syslog(LOG_ERR,"Error reading compression header") ;
                        rfio_close(fd) ;
                        rfio_unlink(filename) ;
                        i = ETIMEDOUT ;
                        exit(i) ;
                    }
                    msg_compress = ( struct mess_compress *) readbuffer ;
#ifndef WORDS_BIGENDIAN
                    msg_compress->datalen = ntohl(msg_compress->datalen) ;
#endif
                    if ( msg_compress->code == DATA_COMPRESS) {
                        compressionon = 1 ;
                    } else {
                        compressionon = 0 ;
                    }
                    datatoreceive = msg_compress->datalen ;
                } else {
                    /*
                    ** No compression just adjust the length to receive
                    */
                    if (buffersizeperstream*1024  <= nbtoget-nbget ) {
                        datatoreceive =  buffersizeperstream*1024 ;
                    } else {
                        datatoreceive = nbtoget-nbget ;
                    }
                }
                /*
                ** Start the data collection
                */
                dataonone = 0 ;
                while ( dataonone < datatoreceive ) {
                    nfds = sysconf(_SC_OPEN_MAX) ;
                    FD_ZERO(&selectmask) ;
                    FD_SET(recsock,&selectmask) ;
                    wait_timer.tv_sec  = DATASOCKTO ;
                    wait_timer.tv_usec = 0 ;
                    if ( (retcode = select(nfds,&selectmask,0,0,&wait_timer) ) == -1 ) {
                        /*
                        ** Select error
                        */
                        i = errno ;
                        syslog(LOG_ERR,"Error select while receiving : %s",strerror(errno)) ;
                        rfio_close(fd) ;
                        rfio_unlink(filename) ;
                        close(recsock) ;
                        exit(i) ;
                    } else if ( retcode == 0 ) {
                        syslog(LOG_ERR,"Time out while receiving") ;
                        rfio_close(fd) ;
                        rfio_unlink(filename) ;
                        i=ETIMEDOUT ;
                        close(recsock) ;
                        exit(i) ;
                    } else {
                        retcode = recv(recsock,&readbuffer[dataonone],datatoreceive-dataonone,0) ;
                        if ( retcode < 0 ) {
                            i = errno ;
                            syslog(LOG_ERR,"Error while receiving : %s",strerror(errno)) ;
                            rfio_close(fd) ;
                            rfio_unlink(filename) ;
                            close(recsock) ;
                            exit(i) ;
                        } else if ( retcode == 0 ) {
                            i = ECONNRESET ;
                            syslog(LOG_ERR,"Connexion breaks") ;
                            rfio_close(fd) ;
                            rfio_unlink(filename) ;
                            close(recsock) ;
                            exit(i) ;
                        } else {
                            dataonone = dataonone + retcode ;
                        }
                    }
                }
                /*
                ** We have received all data needed
                */
                if ( (transferoption & TROPT_GZIP ) == TROPT_GZIP ) {
                    if ( compressionon == 1 ) {
                        bufcomplen = buffersizeperstream*1024 ;
                        buflen = dataonone ;
                        retcode = uncompress((Bytef *) compbuffer, &bufcomplen, (Bytef *) readbuffer, buflen) ;
                        if ( retcode != 0 ) {
                            i = EILSEQ ;
                            syslog(LOG_ERR,"Error while decompressing %d ",retcode) ;
                            rfio_close(fd) ;
                            rfio_unlink(filename) ;
                            close(recsock) ;
                            exit(i) ;
                        }
                        memcpy(readbuffer,compbuffer,buffersizeperstream*1024) ;
                        lentowrite = bufcomplen ;
                    } else {
                        lentowrite = dataonone ;
                    }
                } else {
                    lentowrite = dataonone ;
                }
                /*
                ** Write it to the file
                */
                lenwrited = 0 ;
                while ( lenwrited < lentowrite ) {
                    if ( (retcode = rfio_write(fd,&readbuffer[lenwrited],lentowrite-lenwrited)) < 0 ) {
                        i = errno ;
                        syslog(LOG_ERR,"error writing file : %s",strerror(errno)) ;
                        rfio_close(fd) ;
                        rfio_unlink(filename) ;
                        if ( i == EDQUOT ||
                                i == ENOSPC ) {
                            close(recsock) ;
                            exit(255) ;
                        } else {
                            close(recsock) ;
                            exit(i) ;
                        }
                    } 
                    lenwrited = lenwrited + retcode ;
                }
                nbget = nbget+lenwrited ;
            }
            /*
            ** All data have been received so send the ACK message
            */
            msg = (struct message *) readbuffer ;
            msg->code = MSG_ACK ;
            msg->msglen = 0 ;
            if ( writemessage(recsock,readbuffer,MINMESSLEN,SENDCONTROLTO) < 0 ) {
                rfio_close(fd) ;
                rfio_unlink(filename) ;
                syslog(LOG_ERR,"Error sending ACK ") ;
                close(recsock) ;
                exit(ETIMEDOUT) ;
            }
            toprint64 = nbget ;
            syslog(LOG_NOTICE,"Child received %" LONG_LONG_FORMAT " bytes ; end correct ",toprint64) ;
            rfio_close(fd) ;
            close(recsock) ;
            exit(0) ;
        } else {
            /*
            ** We are in father
            */
            if ( retcode == -1 ) {
                /*
                ** Fork failed ...
                */
                syslog(LOG_ERR,"fork failed : %s",strerror(errno)) ;
                rfio_unlink(filename) ;
                sprintf(logmessage,"fork failed : %s ",strerror(errno)) ;
                if ( childendinerror == 0 ) {
                    childendinerror = 1 ;
                    reply(MSG_BAD,logmessage) ;
                }
                clean_child() ;
                return 1 ;
            } else {
                *pidfree++ = retcode ;
                nbpidchild++ ;
                close(recsock) ;
            }
        }
    }
    /*
    ** Set the state before starting children because if the file was
    ** small the child has ended before state was setup to correct value
    */
    state = S_RECEIVING ;
    /*
    ** Start all children
    */
    pidfree = mychildren ;
    for (i = 0 ; i<nbpidchild ; i++) {
        if ( *pidfree != 0 ) {
            kill(*pidfree,SIGHUP) ;
        }
       pidfree++ ;
    }
#endif 
    return 0 ;
}
