#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <bbftpd.h>
#include <common.h>
#include <config.h>
#include <structures.h>

int  sendwinsize ;		
int  recvwinsize ;		
struct sockaddr_in his_addr;		/* Remote adresse */
struct sockaddr_in ctrl_addr;	/* Local adresse */
int  fixeddataport ;

/*******************************************************************************
** socketcreatereceive :                                                       *
**                                                                             *
**      Routine to create a receive socket                                     *
**      This routine may send MSG_INFO                                         *
**                                                                             *
**      INPUT variable :                                                       *
**          dirname    :  portnumber to connect    NOT MODIFIED                *
**                                                                             *
**      OUPUT variable :                                                       *
**          logmessage :  to write the error message in case of error          *
**                                                                             *
**      RETURN:                                                                *
**          -1   Creation failed unrecoverable error                           *
**           0   OK                                                            *
**           >0  Creation failed recoverable error                             *
**                                                                             *
*******************************************************************************/
int socketcreatereceive(int portnumber,char *logmessage) 
{
	int	sock ;
	struct sockaddr_in data_source ;
	int	on = 1 ;
	int tcpwinsize ;
	int	addrlen ;
	int	retcode ;

	sock = socket ( AF_INET, SOCK_STREAM, IPPROTO_TCP ) ;
	if ( sock < 0 ) {
		syslog(LOG_ERR, "Cannot create receive socket on port %d : %s",portnumber,strerror(errno));
		sprintf(logmessage,"Cannot create receive socket on port %d : %s",portnumber,strerror(errno)) ;
		return (-1) ;
	}
	if ( setsockopt(sock,SOL_SOCKET, SO_REUSEADDR,(char *)&on,sizeof(on)) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot set SO_REUSEADDR on receive socket port %d : %s",portnumber,strerror(errno)) ;
		sprintf(logmessage,"Cannot set SO_REUSEADDR on receive socket port %d : %s",portnumber,strerror(errno)) ;
		return (-1) ;
	}
    tcpwinsize = 1024 * recvwinsize ;
	if ( setsockopt(sock,SOL_SOCKET, SO_RCVBUF,(char *)&tcpwinsize,sizeof(tcpwinsize)) < 0 ) {
        /*
        ** In this case we are going to log a message
        */
		syslog(LOG_ERR,"Cannot set SO_RCVBUF on receive socket port %d : %s",portnumber,strerror(errno)) ;
    }
    addrlen = sizeof(tcpwinsize) ;
    if ( getsockopt(sock,SOL_SOCKET,SO_RCVBUF,(char *)&tcpwinsize,&addrlen) < 0 ) {
 		syslog(LOG_ERR,"Cannot get SO_RCVBUF on receive socket port %d : %s",portnumber,strerror(errno)) ;
		close(sock) ;
		sprintf(logmessage,"Cannot get SO_RCVBUF on receive socket port %d : %s",portnumber,strerror(errno)) ;
		return (-1) ;
    }
    if ( tcpwinsize < 1024 * recvwinsize ) {
        sprintf(logmessage,"Receive buffer on port %d cannot be set to %d Bytes, Value is %d Bytes",portnumber,1024*recvwinsize,tcpwinsize) ;
  		syslog(LOG_ERR,logmessage) ;
/*
        reply(MSG_INFO,logmessage) ;
*/
	}
    tcpwinsize = 1024 * sendwinsize ;
	if ( setsockopt(sock,SOL_SOCKET, SO_SNDBUF,(char *)&tcpwinsize,sizeof(tcpwinsize)) < 0 ) {
        /*
        ** In this case we just log an error
        */
		syslog(LOG_ERR,"Cannot set SO_SNDBUF  on receive socket port %d : %s",portnumber,strerror(errno)) ;
    }
    addrlen = sizeof(tcpwinsize) ;
	if ( getsockopt(sock,SOL_SOCKET, SO_SNDBUF,(char *)&tcpwinsize,&addrlen) < 0 ) {
		syslog(LOG_ERR,"Cannot get SO_SNDBUF on receive socket port %d : %s",portnumber,strerror(errno)) ;
		close(sock) ;
		sprintf(logmessage,"Cannot get SO_SNDBUF on receive socket port %d : %s",portnumber,strerror(errno)) ;
		return (-1) ;
    }
    if ( tcpwinsize < 1024 * sendwinsize ) {
        sprintf(logmessage,"Send buffer on port %d cannot be set to %d Bytes, Value is %d Bytes",portnumber,1024*sendwinsize,tcpwinsize) ;
  		syslog(LOG_ERR,logmessage) ;
/*
        reply(MSG_INFO,logmessage) ;
*/
	}
	if ( setsockopt(sock,IPPROTO_TCP, TCP_NODELAY,(char *)&on,sizeof(on)) < 0 ) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot set TCP_NODELAY on receive socket port %d : %s",portnumber,strerror(errno)) ;
		sprintf(logmessage,"Cannot set TCP_NODELAY on receive socket port %d : %s",portnumber,strerror(errno)) ;
		return (-1) ;
	}
	data_source.sin_family = AF_INET;
#ifdef SUNOS
	data_source.sin_addr = ctrl_addr.sin_addr;
#else
	data_source.sin_addr.s_addr = INADDR_ANY;
#endif
	if ( fixeddataport == 1 ) {
		data_source.sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);
	} else {
		data_source.sin_port = 0;
	}
	if ( bind(sock, (struct sockaddr *) &data_source,sizeof(data_source)) < 0) {
		close(sock) ;
		syslog(LOG_ERR,"Cannot bind on receive socket port %d : %s",portnumber,strerror(errno)) ;
		sprintf(logmessage,"Cannot bind on receive socket port %d : %s",portnumber,strerror(errno)) ;
		return (-1) ;
	}
	data_source.sin_addr = his_addr.sin_addr;
	data_source.sin_port = htons(portnumber);
	addrlen = sizeof(data_source) ;
	retcode = connect(sock,(struct sockaddr *) &data_source,addrlen) ;
	if ( retcode < 0 ) {
		if ( errno == EINTR ) {
			/*
			** In this case we are going to tell the calling program 
			** to retry so we close the socket and set the return code
			** to zero 
			*/
			close(sock) ;
			syslog(LOG_ERR,"Cannot connect receive socket port %d : %s, telling calling to retry ",portnumber,strerror(errno)) ;
			return 0 ;
		} else {
			close(sock) ;
			syslog(LOG_ERR,"Cannot connect receive socket port %d : %s",portnumber,strerror(errno)) ;
			sprintf(logmessage,"Cannot connect receive socket port %d : %s",portnumber,strerror(errno)) ;
			return (-1) ;
		}
	}
	return sock ;
}

main (argc,argv,envp)
    int     argc ;
    char    **argv ;
    char    **envp ;
{
    char logmessage[1024] ;
    int retcode ;
    
    sendwinsize = 1000 ;
    recvwinsize = 1000 ;
    retcode = socketcreatereceive(40000,logmessage) ; 
    printf("retcode = %d\n",retcode) ;
    
 }
   
    
    

